from big_thing_py.utils import *
import subprocess
import asyncio


def validate_broker_address(address: str) -> bool:
    pattern = re.compile(r'^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}:\d+$')
    return bool(pattern.match(address))


class WiFiManager:
    def __init__(self, ssid: str = '', password: str = ''):
        self._ssid = ssid
        self._password = password

    @property
    def ssid(self) -> str:
        return self._ssid

    @ssid.setter
    def ssid(self, ssid: str) -> None:
        self._ssid = ssid

    @property
    def password(self) -> str:
        return self._password

    @password.setter
    def password(self, password: str) -> None:
        self._password = password

    def get_connected_wifi_ssid(self) -> str:
        try:
            result = subprocess.run(
                ["nmcli", "-t", "-f", "ACTIVE,SSID", "device", "wifi"], check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True
            )
            active_connections = [line for line in result.stdout.split('\n') if line.startswith('yes:')]
            return active_connections[0].split(':')[1]
        except subprocess.CalledProcessError as e:
            MXLOG_DEBUG(f"Failed to get connected WiFi: {e}")
        return ''

    async def find_ssid(self, ssid: str, timeout: int = 10) -> bool:
        end_time = asyncio.get_event_loop().time() + timeout
        while True:
            process = await asyncio.create_subprocess_shell(
                "sudo nmcli dev wifi list --rescan yes", stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE
            )
            stdout, stderr = await process.communicate()
            if ssid in stdout.decode():
                MXLOG_DEBUG(f"Found SSID: {ssid}")
                return True
            elif asyncio.get_event_loop().time() > end_time:
                MXLOG_DEBUG("Timeout: SSID not found within the given time.")
                return False
            await asyncio.sleep(1)  # 잠시 대기 후 다시 시도합니다.

    async def connect_to(self, ssid: str, password: str) -> bool:
        try:
            await asyncio.create_subprocess_shell("nmcli --version", stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE)
        except OSError:
            MXLOG_DEBUG("nmcli is not installed. Please install NetworkManager to use this function.")
            return False

        ssid_found = await self.find_ssid(ssid)
        if not ssid_found:
            MXLOG_DEBUG(f"SSID {ssid} not found. Cannot attempt to connect.")
            return False

        connect_cmd = f'sudo nmcli dev wifi connect "{ssid}" password "{password}"'
        try:
            process = await asyncio.create_subprocess_shell(connect_cmd, stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE)
            stdout, stderr = await process.communicate()
            if process.returncode == 0:
                MXLOG_DEBUG("WiFi connection attempt: success")
                return True
            else:
                MXLOG_DEBUG(f"WiFi connection attempt: failed\n{stderr.decode()}")
                return False
        except OSError as e:
            MXLOG_DEBUG(f"Error executing nmcli command: {e}")
            return False

    async def connect(self) -> bool:
        return await self.connect_to(self._ssid, self._password)

    def network_reachable(self, url: str, port: int, retry=3, delay=0.5):
        attempt = 0
        while attempt < retry:
            try:
                socket.create_connection((url, port), timeout=1)
                return True
            except OSError:
                attempt += 1
                if attempt < retry:
                    time.sleep(delay)
        return False

    def check_connection(self) -> bool:
        try:
            result = subprocess.run(
                ["nmcli", "-t", "-f", "ACTIVE,SSID", "device", "wifi"], check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True
            )
            active_connections = [line for line in result.stdout.split('\n') if line.startswith('yes:')]
            if active_connections:
                MXLOG_DEBUG("WiFi connection status: connected")
                return True
            else:
                MXLOG_DEBUG("WiFi connection status: not connected")
                return False
        except subprocess.CalledProcessError:
            MXLOG_DEBUG("Failed to check WiFi connection status")
            return False


if __name__ == '__main__':
    import asyncio

    START_LOGGER()

    async def ble_run():
        from big_thing_py.core.ble_advertiser import BLEAdvertiser

        ble_advertiser = BLEAdvertiser(server_name='MySSIX Thing BLE Server')
        await ble_advertiser.start()
        ssid, pw, broker, error_code = await ble_advertiser.wait_until_wifi_credentials_set()
        MXLOG_DEBUG(f'WiFi credentials set: ssid: {ssid}, pw: {pw}, broker: {broker}, error_code: {error_code}')
        await ble_advertiser.stop()
        return ssid, pw, broker, error_code

    async def wifi_run(ssid, pw):
        wifi_manager = WiFiManager(ssid, pw)
        MXLOG_DEBUG(f'current wifi: {wifi_manager.get_connected_wifi_ssid()}')
        await wifi_manager.connect_to(ssid, pw)
        MXLOG_DEBUG(f'current wifi: {wifi_manager.get_connected_wifi_ssid()}')

    ssid, pw, broker, error_code = asyncio.run(ble_run())
    asyncio.run(wifi_run(ssid, pw))



# from . import Capability
from .Objects import (
    SkillValue,
    SkillFunction,
    SkillFunctionArgument,
    airConditionerMode,
    airPurifierFanMode,
    atmosphericPressureSensor,
    audioMute,
    audioVolume,
    battery,
    blind,
    blindLevel,
    button,
    camera,
    carbonDioxideMeasurement,
    chargingState,
    clock,
    colorControl,
    contactSensor,
    currentMeasurement,
    curtain,
    curtainLevel,
    dustSensor,
    dehumidifierMode,
    dishwasherMode,
    doorControl,
    emailProvider,
    fanControl,
    feederOperatingState,
    feederPortion,
    gasMeter,
    humidifierMode,
    irrigatorOperatingState,
    irrigatorPortion,
    leakSensor,
    lightLevel,
    manager,
    mediaPlayback,
    menuProvider,
    motionSensor,
    pumpOperationMode,
    powerMeter,
    presenceSensor,
    pump,
    refresh,
    relativeHumidityMeasurement,
    robotCleanerCleaningMode,
    sirenMode,
    smokeDetector,
    soilHumidityMeasurement,
    soundPressureLevel,
    soundSensor,
    switch,
    switchLevel,
    temperatureMeasurement,
    testSkill,
    tvChannel,
    tvocMeasurement,
    valve,
    voltageMeasurement,
    weatherProvider,
    windowControl,
    windowShade,
    windowShadeLevel,
)

__all__ = [
    # 'Skill',
    'SkillValue',
    'SkillFunction',
    'SkillFunctionArgument',
    'airConditionerMode',
    'airPurifierFanMode',
    'atmosphericPressureSensor',
    'audioMute',
    'audioVolume',
    'battery',
    'blind',
    'blindLevel',
    'button',
    'camera',
    'carbonDioxideMeasurement',
    'chargingState',
    'clock',
    'colorControl',
    'contactSensor',
    'currentMeasurement',
    'curtain',
    'curtainLevel',
    'dustSensor',
    'dehumidifierMode',
    'dishwasherMode',
    'doorControl',
    'emailProvider',
    'fanControl',
    'feederOperatingState',
    'feederPortion',
    'gasMeter',
    'humidifierMode',
    'irrigatorOperatingState',
    'irrigatorPortion',
    'leakSensor',
    'lightLevel',
    'manager',
    'mediaPlayback',
    'menuProvider',
    'motionSensor',
    'pumpOperationMode',
    'powerMeter',
    'presenceSensor',
    'pump',
    'refresh',
    'relativeHumidityMeasurement',
    'robotCleanerCleaningMode',
    'sirenMode',
    'smokeDetector',
    'soilHumidityMeasurement',
    'soundPressureLevel',
    'soundSensor',
    'switch',
    'switchLevel',
    'temperatureMeasurement',
    'testSkill',
    'tvChannel',
    'tvocMeasurement',
    'valve',
    'voltageMeasurement',
    'weatherProvider',
    'windowControl',
    'windowShade',
    'windowShadeLevel',
]



from big_thing_py.big_thing import *
from big_thing_py.staff_thing import MXStaffThing
from big_thing_py.manager import *
from big_thing_py.core.service_model import Objects as Skill
import uuid
import pwd


class MXManagerThing(MXBigThing, metaclass=ABCMeta):
    DEFAULT_STAFF_THING_STORE_PATH = 'staff_thing_info.json'

    def __init__(
        self,
        name: str = MXThing.DEFAULT_NAME,
        nick_name: str = MXThing.DEFAULT_NAME,
        category=DeviceCategory.ManagerThing,
        desc='',
        version=sdk_version(),
        service_list: List[MXService] = [],
        alive_cycle: float = 60,
        ip: str = '127.0.0.1',
        port: int = 1883,
        ssl_ca_path: str = '',
        ssl_cert_path: str = '',
        ssl_key_path: str = '',
        log_name: str = '',
        log_enable: bool = True,
        log_mode: MXPrintMode = MXPrintMode.ABBR,
        append_mac_address: bool = True,
    ):
        super().__init__(
            name=name,
            nick_name=nick_name,
            category=category,
            desc=desc,
            version=version,
            service_list=service_list,
            alive_cycle=alive_cycle,
            is_super=False,
            is_parallel=True,
            is_builtin=False,
            is_manager=True,
            is_matter=False,
            ip=ip,
            port=port,
            ssl_ca_path=ssl_ca_path,
            ssl_cert_path=ssl_cert_path,
            ssl_key_path=ssl_key_path,
            log_name=log_name,
            log_enable=log_enable,
            log_mode=log_mode,
            append_mac_address=append_mac_address,
        )

        self._staff_thing_list: List[MXStaffThing] = []
        self._subscribe_event_server_task: asyncio.Task = None

    @override
    async def _setup(self):
        default_tag_list = [MXTag('manager')]

        value_list = []
        function_list = [
            MXFunction(
                func=self._discover,
                name='discover',
                category=Skill.manager.Functions.discover,
                tag_list=default_tag_list,
                return_type=MXType.STRING,
                arg_list=[
                    # MXArgument(name='timeout', type=MXType.DOUBLE, bound=(0, 60)),
                ],
                exec_time=30,
                timeout=60,
            ),
            MXFunction(
                func=self._add_thing,
                name='add_thing',
                category=Skill.manager.Functions.add_thing,
                tag_list=default_tag_list,
                return_type=MXType.BOOL,
                arg_list=[
                    MXArgument(name='staff_thing_info', type=MXType.STRING, bound=(0, 999999)),
                    MXArgument(name='client_id', type=MXType.STRING, bound=(0, 999999)),
                ],
                exec_time=30,
                timeout=60,
            ),
            MXFunction(
                func=self._delete_thing,
                name='delete_thing',
                category=Skill.manager.Functions.delete_thing,
                tag_list=default_tag_list,
                return_type=MXType.BOOL,
                arg_list=[
                    MXArgument(name='staff_thing_info', type=MXType.STRING, bound=(0, 999999)),
                    MXArgument(name='client_id', type=MXType.STRING, bound=(0, 999999)),
                ],
                exec_time=30,
                timeout=60,
            ),
        ]

        await super()._setup()
        self._subscribe_event_server_task = asyncio.create_task(self._subscribe_event_server())

        for service in value_list + function_list:
            self.add_service(service)

    @override
    async def _wrapup(self):
        try:
            if self._mqtt_client and self._mqtt_client.is_connected:
                self._send_TM_UNREGISTER(self._thing_data)
                for thing in self._staff_thing_list:
                    self._send_TM_UNREGISTER(thing)
                # FIXME: Need to wait for the result of unregister
                # recv_msg = await self._receive_queue[MXProtocolType.Base.MT_RESULT_UNREGISTER].get()
                # error = self._handle_mqtt_message(recv_msg, target_thing=self._thing_data)

            return True
        except Exception as e:
            print_error(e)
            return False
        finally:
            if await self._ble_advertiser.is_advertising():
                await self._ble_advertiser.stop()

            self._unregister_mdns_service()

            if self._mqtt_client and self._mqtt_client.is_connected:
                await self._disconnect_from_broker(disconnect_try=MXBigThing.CONNECT_RETRY)

            await self._close_event_server()

            MXLOG_DEBUG('Thing Exit', 'red')
            sys.exit(0)

    # ===========================================================================================
    #  _    _                             _    __                      _    _
    # | |  | |                           | |  / _|                    | |  (_)
    # | |_ | |__   _ __   ___   __ _   __| | | |_  _   _  _ __    ___ | |_  _   ___   _ __   ___
    # | __|| '_ \ | '__| / _ \ / _` | / _` | |  _|| | | || '_ \  / __|| __|| | / _ \ | '_ \ / __|
    # | |_ | | | || |   |  __/| (_| || (_| | | |  | |_| || | | || (__ | |_ | || (_) || | | |\__ \
    #  \__||_| |_||_|    \___| \__,_| \__,_| |_|   \__,_||_| |_| \___| \__||_| \___/ |_| |_||___/
    # ===========================================================================================

    @override
    async def _preprocessing(self):
        await asyncio.sleep(THREAD_TIME_OUT)

        if self._execute_task_list:
            tasks = await asyncio.wait(self._execute_task_list, return_when=asyncio.FIRST_COMPLETED)
            done_tasks: List[asyncio.Task] = tasks[0]
            for task in done_tasks:
                exception = task.exception()
                if exception:
                    print_error(exception)
                else:
                    execute_request: MXExecuteRequest = task.result()

                    error_list = []
                    for thing in self._whole_thing_list:
                        if find_class_in_hierarchy(thing, MXManagerThing) or type(thing) is MXThing:
                            target_thing = self._thing_data
                        elif find_class_in_hierarchy(thing, MXStaffThing):
                            target_thing = thing
                        else:
                            raise ValueError(f'Unknown type instance: {type(thing)}')

                        if target_thing.name != execute_request.result_msg.thing_name:
                            continue

                        error = self._send_TM_RESULT_EXECUTE(execute_request, target_thing)
                        error_list.append(error)
                        if error != MXErrorCode.NO_ERROR:
                            continue

                        dependency_table = target_thing.dependency_table
                        dependency_info_list = dependency_table.get(execute_request.result_msg.function_name, [])
                        for dependency_info in dependency_info_list:
                            value: MXValue = dependency_info['value']
                            if dependency_info['set_value_as_result']:
                                target_function = thing.get_function(execute_request.result_msg.function_name)
                                set_value: Union[int, float, bool, str] = target_function.return_value
                            else:
                                set_value: Union[int, float, bool, str] = dependency_info['set_value']

                            new_value = await value.update(set_value=set_value)
                            if new_value is None:
                                continue

                            self._send_TM_VALUE_PUBLISH(value)

                            set_value_with_cond = dependency_info['set_value_with_cond']
                            if set_value_with_cond:
                                set_value = set_value_with_cond(new_value)
                            else:
                                set_value: Union[int, float, bool, str] = dependency_info['set_value']
                                new_value = await value.update(set_value=set_value)
                                if new_value is None:
                                    continue

                    if all([error == MXErrorCode.NOT_FOUND_ERROR for error in error_list]):
                        MXLOG_DEBUG(f'[{get_current_function_name()}] Target function not found!!!', 'red')
                    elif any([not error in MXBigThing.HANDLE_MQTT_MESSAGE_IGNORE_ERROR_LIST for error in error_list]):
                        MXLOG_DEBUG(f'[{get_current_function_name()}] Send function result failed!!!', 'red')

            self._execute_task_list = [task for task in self._execute_task_list if not task.done()]

    @override
    async def _BROKER_CONNECTED_state_process(self):
        # Prepare to register Manager Thing
        self._subscribe_init_topic_list(self._thing_data)
        if self.is_builtin or self.is_manager:
            MXLOG_DEBUG(f'Run builtin or plugin Thing', 'yellow')
            self._send_TM_REGISTER(self._thing_data)

            recv_msg = await self._receive_queue[MXProtocolType.Base.MT_RESULT_REGISTER].get()
            self._handle_mqtt_message(recv_msg, target_thing=self._thing_data)

        # Prepare to register Staff Things
        loaded_staff_thing_list = await self._load_staff_thing_info()
        for loaded_staff_thing in loaded_staff_thing_list:
            latest_staff_thing = await self._check_thing_exist(loaded_staff_thing)
            if not latest_staff_thing:
                MXLOG_WARN(f'[{get_current_function_name()}] Staff Thing {loaded_staff_thing.name} is offline', 'red')
                loaded_staff_thing.is_alive = False
                continue

            latest_staff_thing.is_alive = True
            latest_staff_thing.last_alive_time = get_current_datetime()
            if latest_staff_thing.name == loaded_staff_thing.name and latest_staff_thing.nick_name == loaded_staff_thing.nick_name:
                MXLOG_DEBUG(f'Load staff thing {loaded_staff_thing.name} completed!', 'cyan')
            else:
                MXLOG_DEBUG(
                    f'Update staff thing\'s nick name {loaded_staff_thing.name} completed! {loaded_staff_thing.nick_name} -> {latest_staff_thing.nick_name}',
                    'cyan',
                )
                self._delete_staff_thing_info(loaded_staff_thing)
                self._store_staff_thing_info(latest_staff_thing)
            await self._add_staff_thing(latest_staff_thing)

            self._subscribe_init_topic_list(latest_staff_thing)
            if self.is_builtin or self.is_manager:
                MXLOG_DEBUG(f'Run builtin or plugin Thing', 'yellow')
                self._send_TM_REGISTER(latest_staff_thing)

                recv_msg = await self._receive_queue[MXProtocolType.Base.MT_RESULT_REGISTER].get()
                self._handle_mqtt_message(recv_msg, target_thing=latest_staff_thing, state_change=False)

    @override
    async def _RUNNING_state_process(self):
        # MQTT receive handling
        if not self._receive_queue_empty():
            recv_msg = await self._receive_queue_get()

            # # MQTT receive handling(Manager Thing)
            # error = self._handle_mqtt_message(recv_msg, target_thing=self._thing_data)
            # if error == MXErrorCode.INVALID_DESTINATION:

            # if not error in MXBigThing.HANDLE_MQTT_MESSAGE_IGNORE_ERROR_LIST:
            #     MXLOG_DEBUG(f'[{get_current_function_name()}] MQTT Message handling failed', 'red')

            # MQTT receive handling(Staff Thing)
            for thing in self._whole_thing_list:
                if isinstance(thing, MXManagerThing):
                    thing_data = thing._thing_data
                    state_change = True
                elif isinstance(thing, MXStaffThing):
                    thing_data = thing
                    state_change = False

                target_thing_name = topic_split(decode_MQTT_message(recv_msg)[0])[-1]
                if target_thing_name != thing_data.name:
                    continue

                error = self._handle_mqtt_message(recv_msg, target_thing=thing_data, state_change=state_change)
                if not error in MXBigThing.HANDLE_MQTT_MESSAGE_IGNORE_ERROR_LIST:
                    MXLOG_DEBUG(f'[{get_current_function_name()}] MQTT Message handling failed', 'red')

        # Value publish & Alive handling (Manager Thing)
        current_time = get_current_datetime()
        for value in self.value_list:
            if not value.is_initialized:
                await value.update()
                self._send_TM_VALUE_PUBLISH(value)
            # NOTE (thsvkd): cycle == 0 mean, value is event-based
            elif current_time - value.last_update_time > value.cycle and value.cycle != 0:
                new_value = await value.update()
                if new_value is None:
                    continue

                self._send_TM_VALUE_PUBLISH(value)

        if current_time - self.last_alive_time > self.alive_cycle / MXBigThing.ALIVE_CYCLE_SCALER:
            self._send_TM_ALIVE(self._thing_data)

        # Value publish & Alive handling (Staff Thing)
        for thing in self._staff_thing_list:
            for value in thing.value_list:
                if not thing.is_alive:
                    continue

                if not value.is_initialized:
                    await value.update()
                    self._send_TM_VALUE_PUBLISH(value)
                # NOTE (thsvkd): cycle == 0 mean, value is event-based
                elif current_time - value.last_update_time > value.cycle and value.cycle != 0:
                    new_value = await value.update()
                    if new_value is None:
                        continue

                    self._send_TM_VALUE_PUBLISH(value)

            if current_time - thing.last_alive_time > thing.alive_cycle / MXBigThing.ALIVE_CYCLE_SCALER:
                if await self._check_thing_exist(thing):
                    if not thing.is_alive:
                        MXLOG_DEBUG(f'Staff Thing {thing.name} is online', 'cyan')
                        thing.is_alive = True
                        self._send_TM_REGISTER(thing)

                    self._send_TM_ALIVE(thing)
                else:
                    if thing.is_alive:
                        MXLOG_DEBUG(f'Staff Thing {thing.name} is offline', 'red')
                        thing.is_alive = False
                        self._send_TM_UNREGISTER(thing)

    @override
    async def _BROKER_RECONNECTED_state_process(self):
        get_home_msg = self._thing_data.generate_get_home_message().mqtt_message()
        result_home_topic = MXProtocolType.WebClient.ME_RESULT_HOME.value % '+'
        self._subscribe(result_home_topic)

        current_time = get_current_datetime()
        if get_current_datetime() - current_time < MXBigThing.MIDDLEWARE_RECONNECT_TIMEOUT:
            self._publish(get_home_msg.topic, get_home_msg.payload)
            try:
                recv_msg = self._receive_queue[MXProtocolType.WebClient.ME_RESULT_HOME].get_nowait()
                self._handle_mqtt_message(recv_msg, target_thing=self._thing_data)

                # Auto subscription restore feature
                for topic in self.subscribed_topic_set:
                    self._subscribe(topic)

                self._send_TM_ALIVE(self._thing_data)
                for thing in self._staff_thing_list:
                    self._send_TM_ALIVE(thing)
            except asyncio.QueueEmpty:
                time.sleep(MXBigThing.MIDDLEWARE_ONLINE_CHECK_INTERVAL)
        else:
            if self.is_builtin or self.is_manager:
                self.next_state = ThingState.SHUTDOWN
            else:
                MXLOG_DEBUG(f'Middleware is offline... Go back to BLE setup.', 'red')
                self.next_state = ThingState.BLE_ADVERTISE

    # ======================================================================================================================= #
    #  _    _                    _  _         __  __   ____  _______  _______   __  __                                        #
    # | |  | |                  | || |       |  \/  | / __ \|__   __||__   __| |  \/  |                                       #
    # | |__| |  __ _  _ __    __| || |  ___  | \  / || |  | |  | |      | |    | \  / |  ___  ___  ___   __ _   __ _   ___    #
    # |  __  | / _` || '_ \  / _` || | / _ \ | |\/| || |  | |  | |      | |    | |\/| | / _ \/ __|/ __| / _` | / _` | / _ \   #
    # | |  | || (_| || | | || (_| || ||  __/ | |  | || |__| |  | |      | |    | |  | ||  __/\__ \\__ \| (_| || (_| ||  __/   #
    # |_|  |_| \__,_||_| |_| \__,_||_| \___| |_|  |_| \___\_\  |_|      |_|    |_|  |_| \___||___/|___/ \__,_| \__, | \___|   #
    #                                                                                                         __/ |           #
    #                                                                                                         |___/           #
    # ======================================================================================================================= #

    # ===========================
    #            _____   _____
    #     /\    |  __ \ |_   _|
    #    /  \   | |__) |  | |
    #   / /\ \  |  ___/   | |
    #  / ____ \ | |      _| |_
    # /_/    \_\|_|     |_____|
    # ===========================

    @abstractmethod
    def _extract_staff_thing_info(self, staff_thing_info: dict) -> dict:
        '''
        staff thing 정보를 추출하는 함수

        Args:
            staff_thing_info (dict): staff thing의 정보를 담고 있는 딕셔너리

        Returns:
            dict: 추출된 staff thing 정보를 담고 있는 딕셔너리
        '''
        pass

    @abstractmethod
    def _create_staff(self, staff_thing_info: dict) -> Union[MXStaffThing, None]:
        '''
        _scan_staff_thing() 함수를 통해 수집된 staff thing 정보를 바탕으로 staff thing을 생성하는 함수.
        만약 스캔하는 것만으로 완벽한 staff thing의 정보를 수집할 수 없다면, staff thing의 register 메시지를 받아 처리하는
        _handle_REGISTER_staff_message() 함수에서 staff thing을 self._staff_thing_list에서 찾아 정보를 추가할 수 있다.

        Args:
            staff_thing_info (dict): staff thing의 정보를 담고 있는 딕셔너리

        Returns:
            staff_thing(MXStaffThing): 생성한 staff thing 인스턴스
        '''
        pass

    @abstractmethod
    async def _scan_staff_thing(self, timeout: float) -> Tuple[List[dict], str]:
        '''
        지속적으로 staff thing을 발견하여 정보를 수집하여 반환하는 함수.
        timeout을 지정하여 한 번 staff thing을 검색하는데 소요될 시간을 지정할 수 있다.

        Args:
            timeout (float): staff thing을 검색하는데 소요될 시간

        Returns:
            List[dict]: staff thing의 정보를 담고 있는 리스트
        '''
        pass

    @abstractmethod
    async def _check_thing_exist(self, staff_thing: MXStaffThing) -> Union[bool, MXStaffThing]:
        '''
        staff thing이 실제로 존재하는 지 검사하는 함수

        Args:
            staff_thing (MXStaffThing): staff thing의 정보를 담고 있는 문자열

        Returns:
            Union[bool, MXStaffThing]: staff thing가 존재하는 경우 staff thing 인스턴수를, 존재하지 않는 경우 False를 반환
        '''
        pass

    @abstractmethod
    async def _add_thing__(self, staff_thing_info_string: str, client_id: str) -> Union[MXStaffThing, None]:
        '''
        staff thing을 추가하는 함수.

        Args:
            staff_thing_info_string (str): staff thing의 정보를 담고 있는 문자열
            client_id (str): staff thing을 등록 요청하는 클라이언트의 ID

        Returns:
            dict: staff thing 추가 결과 딕셔너리
        '''
        pass

    @abstractmethod
    async def _delete_thing__(self, staff_thing_info_string: str, client_id: str) -> Union[MXStaffThing, None]:
        '''
        staff thing을 삭제하는 함수.

        Args:
            staff_thing_info_string (str): staff thing의 정보를 담고 있는 문자열
            client_id (str): staff thing을 삭제 요청하는 클라이언트의 ID

        Returns:
            dict: staff thing 삭제 결과 딕셔너리
        '''
        pass

    @abstractmethod
    async def _subscribe_event_server(self) -> asyncio.Task:
        '''
        event server를 구독하는 함수

        Args: None

        Returns:
            asyncio.Task: event server를 구독하는 task
        '''
        pass

    @abstractmethod
    async def _close_event_server(self) -> None:
        '''
        event server를 종료하는 함수

        Args: None

        Returns: None
        '''
        pass

    async def _add_thing(self, staff_thing_info_string: str, client_id: str) -> dict:
        target_staff_thing = await self._add_thing__(staff_thing_info_string, client_id)
        if target_staff_thing:
            self._store_staff_thing_info(target_staff_thing)
            return dict(thing=target_staff_thing.name, error=MXErrorCode.NO_ERROR.value)
        else:
            return dict(thing='', error=MXErrorCode.FAIL.value)

    async def _delete_thing(self, staff_thing_info_string: str, client_id: str) -> dict:
        target_staff_thing = await self._delete_thing__(staff_thing_info_string, client_id)
        if target_staff_thing:
            self._delete_staff_thing_info(target_staff_thing)
            return dict(thing=target_staff_thing.name, error=MXErrorCode.NO_ERROR.value)
        else:
            return dict(thing='', error=MXErrorCode.FAIL.value)

    async def _discover(self) -> str:
        staff_thing_list, error_string = await self._scan_staff_thing()
        if staff_thing_list:
            discovered_things = dict(things=staff_thing_list, error=MXErrorCode.NO_ERROR.value, error_string='')
        else:
            discovered_things = dict(things=[], error=MXErrorCode.FAIL.value, error_string=error_string)

        return dict_to_json_string(discovered_things)

    # ===============
    # ___  ___ _____
    # |  \/  ||_   _|
    # | .  . |  | |
    # | |\/| |  | |
    # | |  | |  | |
    # \_|  |_/  \_/
    # ===============

    # ===============
    #  _____ ___  ___
    # |_   _||  \/  |
    #   | |  | .  . |
    #   | |  | |\/| |
    #   | |  | |  | |
    #   \_/  \_|  |_/
    # ===============

    # ========================
    #         _    _  _
    #        | |  (_)| |
    #  _   _ | |_  _ | | ___
    # | | | || __|| || |/ __|
    # | |_| || |_ | || |\__ \
    #  \__,_| \__||_||_||___/
    # ========================

    def _get_thing(self, thing_name: str) -> MXThing:
        for thing in self._whole_thing_list:
            if thing.name == thing_name:
                return thing

    async def _add_staff_thing(self, staff_thing: MXStaffThing) -> MXStaffThing:
        if staff_thing in self._staff_thing_list:
            MXLOG_WARN(f'[{get_current_function_name()}] Staff Thing {staff_thing.name} is already added')
        else:
            if not await staff_thing.setup():
                raise ValueError(f'[{get_current_function_name()}] Staff Thing {staff_thing.name} setup failed')

            staff_thing.is_alive = True
            self._staff_thing_list.append(staff_thing)

        return staff_thing

    async def _delete_staff_thing(self, staff_thing: MXStaffThing) -> bool:
        if not staff_thing in self._staff_thing_list:
            MXLOG_WARN(f'[{get_current_function_name()}] Staff Thing {staff_thing.name} is not exist')

        if not await staff_thing.wrapup():
            raise ValueError(f'[{get_current_function_name()}] Staff Thing {staff_thing.name} wrapup failed')

        self._staff_thing_list.remove(staff_thing)
        return True

    def _store_staff_thing_info(self, target_staff_thing: MXStaffThing) -> None:
        if not os.path.exists(self.DEFAULT_STAFF_THING_STORE_PATH):
            json_file_write(self.DEFAULT_STAFF_THING_STORE_PATH, dict(data=[]), mode='w')

        if os.geteuid() == 0:
            original_user = os.environ.get('SUDO_USER')
            if not original_user:
                raise ValueError('SUDO_USER environment variable is not set. Unable to determine the original user.')

            user_info = pwd.getpwnam(original_user)
            uid = user_info.pw_uid
            gid = user_info.pw_gid

            os.chown(self.DEFAULT_STAFF_THING_STORE_PATH, uid, gid)
            os.chmod(self.DEFAULT_STAFF_THING_STORE_PATH, 0o644)

        staff_thing_info_list = []
        staff_thing_info_file: Dict[str, list] = json_file_read(self.DEFAULT_STAFF_THING_STORE_PATH)
        staff_thing_info_list = staff_thing_info_file['data']

        if not target_staff_thing.staff_dict() in staff_thing_info_list:
            staff_thing_info_list.append(target_staff_thing.staff_dict())

        json_file_write(self.DEFAULT_STAFF_THING_STORE_PATH, dict(data=staff_thing_info_list))

    def _delete_staff_thing_info(self, target_staff_thing: MXStaffThing) -> None:
        staff_thing_info_file = json_file_read(self.DEFAULT_STAFF_THING_STORE_PATH)
        staff_thing_info_list = staff_thing_info_file['data']
        staff_thing_info_list = [info for info in staff_thing_info_list if info['id'] != target_staff_thing.name and info['name'] != target_staff_thing.nick_name]
        json_file_write(self.DEFAULT_STAFF_THING_STORE_PATH, dict(data=staff_thing_info_list))

    async def _load_staff_thing_info(self) -> List[MXStaffThing]:
        staff_thing_info_list = json_file_read(self.DEFAULT_STAFF_THING_STORE_PATH)
        if not staff_thing_info_list:
            return []
        else:
            staff_thing_info_list = [self._create_staff(staff_thing_info) for staff_thing_info in staff_thing_info_list['data']]

        return staff_thing_info_list

    @override
    def _subscribe_init_topic_list(self, staff_thing: MXThing) -> None:
        topic_list = [
            MXProtocolType.Base.MT_REQUEST_REGISTER_INFO.value % staff_thing.name,
            MXProtocolType.Base.MT_RESULT_REGISTER.value % staff_thing.name,
            MXProtocolType.Base.MT_RESULT_UNREGISTER.value % staff_thing.name,
            MXProtocolType.Base.MT_RESULT_BINARY_VALUE.value % staff_thing.name,
        ]

        for topic in topic_list:
            self._subscribe(topic)

    @override
    def _subscribe_service_topic_list(self, thing: MXThing):
        topic_list = []

        if find_class_in_hierarchy(thing, MXManagerThing) or type(thing) is MXThing:
            target_thing = self._thing_data
            for service in target_thing.function_list:
                topic_list += [
                    MXProtocolType.Base.MT_IN_EXECUTE.value % (service.name, target_thing.name),
                ]
        elif find_class_in_hierarchy(thing, MXStaffThing):
            target_thing = thing
            for function in target_thing.function_list:
                topic_list += [
                    MXProtocolType.Base.MT_EXECUTE.value % (function.name, target_thing.name, '+', '+'),
                    (MXProtocolType.Base.MT_EXECUTE.value % (function.name, target_thing.name, '', '')).rstrip('/'),
                ]
        else:
            raise ValueError(f'Unknown type instance: {type(thing)}')

        for topic in topic_list:
            self._subscribe(topic)

    @override
    def _unsubscribe_all_topic_list(self, staff_thing: MXThing):
        # whenever _unsubscribe function execute, it remove target topic from self._subscribed_topic_set
        # so it need to iterate with copy of self._subscribed_topic_set
        topic_list = list(staff_thing.subscribed_topic_set)

        for topic in topic_list:
            self._unsubscribe(topic)

    def _get_staff_thing_by_name(self, name: str, staff_thing_pool: List[MXStaffThing] = None) -> MXStaffThing:
        if staff_thing_pool is None:
            staff_thing_pool = self._staff_thing_list

        for staff_thing in staff_thing_pool:
            if staff_thing.name == name:
                return staff_thing

        return False

    # ====================================
    #               _    _
    #              | |  | |
    #   __ _   ___ | |_ | |_   ___  _ __
    #  / _` | / _ \| __|| __| / _ \| '__|
    # | (_| ||  __/| |_ | |_ |  __/| |
    #  \__, | \___| \__| \__| \___||_|
    #   __/ |
    #  |___/
    # ====================================

    @property
    def _whole_thing_list(self) -> List[MXThing]:
        return [self, *self._staff_thing_list]



from big_thing_py.core.service_model import SkillObjects
from big_thing_py.core.service_model.SkillObjects import MXClassproperty
from typing import Set, Dict, List, Union, Any

ALL_DEVICE_TYPES: dict[int, type["MXDeviceCategory"]] = {}


class MXDeviceCategory:
    """Base class for MX device types."""

    device_type: int
    skills: Set[SkillObjects.Skill]

    def __init_subclass__(cls, *, device_type: int, **kwargs: Any) -> None:
        """Register a subclass."""
        super().__init_subclass__(**kwargs)
        cls.device_type = device_type
        ALL_DEVICE_TYPES[device_type] = cls

    def __hash__(self) -> int:
        """Return unique hash for this object."""
        return self.device_type

    @MXClassproperty
    def name(cls) -> str:
        """Return the class name as a string."""
        return cls.__name__



from big_thing_py.utils import *
from big_thing_py.core.mqtt_message import *


class MXScheduleStatus(Enum):
    def _generate_next_value_(name: str, start, count, last_values):
        return name.lower()

    INIT = auto()
    CHECK = auto()
    SELECT = auto()
    CONFIRM = auto()
    UNDEFINED = auto()

    @classmethod
    def get(cls, name: str) -> 'MXScheduleStatus':
        try:
            return cls[name.upper()]
        except Exception:
            return cls.UNDEFINED


class MXSuperRefreshMessage(MXMQTTSendMessage):
    def __init__(self, thing) -> None:
        from big_thing_py.core.thing import MXThing

        thing: MXThing = thing

        super().__init__()
        self.protocol_type = MXProtocolType.Super.SM_REFRESH
        self.topic = self.protocol_type.value % (thing.name)
        self.payload = EMPTY_JSON


class MXSuperServiceListResultMessage(MXMQTTReceiveMessage):
    def __init__(self, msg: MQTTMessage) -> None:
        super().__init__(msg=msg)
        self.protocol_type = MXProtocolType.Super.MS_RESULT_SERVICE_LIST
        self.super_thing_name: str = self.topic.split('/')[3]
        self.service_list: List[dict] = self.payload['services']


class MXSuperScheduleMessage(MXMQTTReceiveMessage):
    def __init__(self, msg: MQTTMessage) -> None:
        super().__init__(msg=msg)
        self.protocol_type = MXProtocolType.Super.MS_SCHEDULE

        # topic
        self.super_service_name = self.topic.split('/')[2]
        self.super_thing_name = self.topic.split('/')[3]
        self.super_middleware_name = self.topic.split('/')[4]
        self.requester_middleware_name = self.topic.split('/')[5]

        # payload
        self.scenario = self.payload['scenario']
        self.period = self.payload['period']

        if not isinstance(self.scenario, str) or not isinstance(self.period, (int, float)):
            self.payload_error = True


class MXSuperScheduleResultMessage(MXMQTTSendMessage):
    def __init__(
        self,
        super_service_name: str,
        super_thing_name: str,
        super_middleware_name: str,
        requester_middleware_name: str,
        scenario: str,
        error: MXErrorCode = MXErrorCode.UNDEFINED,
    ) -> None:
        super().__init__()
        self.protocol_type = MXProtocolType.Super.SM_RESULT_SCHEDULE

        self.topic = self.protocol_type.value % (
            super_service_name,
            super_thing_name,
            super_middleware_name,
            requester_middleware_name,
        )
        self.payload = dict(scenario=scenario, error=error.value)


class MXSubScheduleMessage(MXMQTTSendMessage):
    def __init__(
        self,
        sub_service_name: str,
        target_thing_name: str,
        target_middleware_name: str,
        requester_middleware_name: str = None,
        super_thing_name: str = None,
        super_service_name: str = None,
        sub_service_request_order: int = None,
        scenario: str = None,
        period: float = None,
        tag_list: List[str] = [],
        range_type: MXRangeType = None,
        request_ID: str = None,
        status: MXScheduleStatus = None,
    ) -> None:
        super().__init__()
        self.protocol_type = MXProtocolType.Super.SM_SCHEDULE

        # topic
        self.sub_service_name = sub_service_name
        self.target_thing_name = target_thing_name
        self.target_middleware_name = target_middleware_name

        self.requester_middleware_name = requester_middleware_name
        self.super_thing_name = super_thing_name
        self.super_service_name = super_service_name
        self.sub_service_request_order = sub_service_request_order

        self.scenario = scenario
        self.period = period
        self.tag_list = tag_list
        self.range_type = range_type

        self.request_ID = request_ID
        self._status = status

        if not self.request_ID:
            self.request_ID = make_request_ID(
                self.requester_middleware_name,
                self.super_thing_name,
                self.super_service_name,
                self.sub_service_request_order,
            )

        self.topic = self.protocol_type.value % (self.sub_service_name, self.target_middleware_name, self.request_ID)
        self.payload = dict(
            scenario=self.scenario,
            period=self.period,
            status=self._status.value,
            tag_list=self.tag_list,
            range=self.range_type.value,
        )

    @property
    def status(self) -> dict:
        return self._status

    @status.setter
    def status(self, status: MXScheduleStatus):
        if not isinstance(status, MXScheduleStatus):
            raise MXTypeError("status must be an MXScheduleStatus")
        self._status = status
        self.payload = dict(
            scenario=self.scenario,
            period=self.period,
            status=self._status.value,
            tag_list=self.tag_list,
            range=self.range_type.value,
        )


class MXSubScheduleResultMessage(MXMQTTReceiveMessage):
    def __init__(self, msg: MQTTMessage) -> None:
        super().__init__(msg=msg)
        self.protocol_type = MXProtocolType.Super.MS_RESULT_SCHEDULE

        # topic
        self.sub_service_name = self.topic.split('/')[3]
        self.target_thing_name = self.topic.split('/')[4]
        self.target_middleware_name = self.topic.split('/')[5]

        self.request_ID = self.topic.split('/')[6]
        self.requester_middleware_name = self.request_ID.split('@')[0]
        self.super_thing_name = self.request_ID.split('@')[1]
        self.super_service_name = self.request_ID.split('@')[2]
        self.sub_service_request_order = self.request_ID.split('@')[3]

        # payload
        self.scenario = self.payload['scenario']
        self.error = MXErrorCode.get(self.payload['error'])
        self.status = MXScheduleStatus.get(self.payload.get('status', None))  # 'check' or 'confirm'


class MXSuperExecuteMessage(MXMQTTReceiveMessage):
    def __init__(self, msg: MQTTMessage) -> None:
        super().__init__(msg=msg)
        self.protocol_type = MXProtocolType.Super.MS_EXECUTE

        # topic
        self.super_service_name = self.topic.split('/')[2]
        self.super_thing_name = self.topic.split('/')[3]
        self.super_middleware_name = self.topic.split('/')[4]
        self.requester_middleware_name = self.topic.split('/')[5]

        # payload
        self.scenario: str = self.payload.get('scenario', None)
        self.arguments: List[dict] = self.payload.get('arguments', None)

        if not self.scenario or self.arguments == None:
            self.payload_error = True
        elif not all([isinstance(arg.get('order', None), int) for arg in self.arguments]) or not all(
            [isinstance(arg.get('value', None), (int, float, str, bool)) for arg in self.arguments]
        ):
            self.payload_error = True

    def tuple_arguments(self) -> tuple:
        sorted_arguments = sorted(self.arguments, key=lambda x: int(x['order']))
        real_arguments = tuple([argument['value'] for argument in sorted_arguments])
        return real_arguments

    def dict_arguments(self) -> List[dict]:
        self.arguments = sorted(self.arguments, key=lambda x: int(x['order']))
        json_arguments = [dict(order=arg['order'], value=arg['value']) for arg in self.arguments]
        return json_arguments


class MXSuperExecuteResultMessage(MXMQTTSendMessage):
    def __init__(
        self,
        super_service_name: str,
        super_thing_name: str,
        super_middleware_name: str,
        requester_middleware_name: str,
        scenario: str,
        return_type: MXType,
        return_value: Union[int, float, bool, str] = None,
        error: MXErrorCode = MXErrorCode.UNDEFINED,
    ) -> None:
        super().__init__()
        self.protocol_type = MXProtocolType.Super.SM_RESULT_EXECUTE

        # payload
        self.scenario = scenario
        self.return_type = return_type
        self.return_value = return_value
        self.error = error

        self.topic = self.protocol_type.value % (
            super_service_name,
            super_thing_name,
            super_middleware_name,
            requester_middleware_name,
        )
        self.payload = dict(scenario=scenario, return_type=return_type.value, return_value=return_value, error=error.value)


class MXSubExecuteMessage(MXMQTTSendMessage):
    def __init__(
        self,
        sub_service_name: str,
        target_thing_name: str,
        target_middleware_name: str,
        requester_middleware_name: str = None,
        super_thing_name: str = None,
        super_service_name: str = None,
        sub_service_request_order: int = None,
        scenario: str = None,
        arguments: List[dict] = None,
        request_ID: str = None,
    ) -> None:
        super().__init__()
        self.protocol_type = MXProtocolType.Super.SM_EXECUTE

        # topic
        self.sub_service_name = sub_service_name
        self.target_thing_name = target_thing_name
        self.target_middleware_name = target_middleware_name

        self.requester_middleware_name = requester_middleware_name
        self.super_thing_name = super_thing_name
        self.super_service_name = super_service_name
        self.sub_service_request_order = sub_service_request_order

        self.scenario = scenario
        self._arguments = arguments

        self.request_ID = request_ID

        if not self.request_ID:
            self.request_ID = make_request_ID(
                self.requester_middleware_name,
                self.super_thing_name,
                self.super_service_name,
                self.sub_service_request_order,
            )

        self.topic = self.protocol_type.value % (self.sub_service_name, self.target_middleware_name, self.request_ID)
        self.payload = dict(scenario=self.scenario, arguments=self._arguments)

    @property
    def arguments(self) -> dict:
        return self._arguments

    @arguments.setter
    def arguments(self, arguments: tuple):
        if not isinstance(arguments, tuple):
            raise MXTypeError("arguments must be an tuple")
        self._arguments = arguments
        dict_arguments = [dict(order=i, value=arg) for i, arg in enumerate(self._arguments)]
        self.payload = dict(scenario=self.scenario, arguments=dict_arguments)

    # def json_arguments(self) -> List[dict]:
    #     self._arguments = sorted(self._arguments, key=lambda x: int(x['order']))
    #     json_arguments = [dict(order=arg['order'], value=arg['value']) for arg in self._arguments]
    #     return json_arguments


class MXSubExecuteResultMessage(MXMQTTReceiveMessage):
    def __init__(self, msg: MQTTMessage) -> None:
        super().__init__(msg=msg)
        self.protocol_type = MXProtocolType.Super.MS_RESULT_EXECUTE

        # topic
        self.sub_service_name = self.topic.split('/')[3]
        self.target_thing_name = self.topic.split('/')[4]
        self.target_middleware_name = self.topic.split('/')[5]

        self.request_ID = self.topic.split('/')[6]
        self.requester_middleware_name = self.request_ID.split('@')[0]
        self.super_thing_name = self.request_ID.split('@')[1]
        self.super_service_name = self.request_ID.split('@')[2]
        self.sub_service_request_order = self.request_ID.split('@')[3]

        # payload
        self.scenario = self.payload['scenario']
        self.return_type = self.payload['return_type']
        # TODO: 추후에 return_value -> return_values로 변경
        self.return_value = self.payload['return_value']
        self.error = MXErrorCode.get(self.payload['error'])

    def tuple_arguments(self) -> tuple:
        sorted_arguments = sorted(self._arguments, key=lambda x: int(x['order']))
        tuple_arguments = tuple([argument['value'] for argument in sorted_arguments])
        return tuple_arguments

    def dict_arguments(self) -> List[dict]:
        sorted_arguments = sorted(self._arguments, key=lambda x: int(x['order']))
        json_arguments = [dict(order=arg['order'], value=arg['value']) for arg in sorted_arguments]
        return json_arguments



from aenum import auto as aauto
from dataclasses import dataclass, field
import textwrap
import sys
from typing import List, Dict, Any, Tuple, Union

from .enum import MXIntEnum, MXStrEnum
from .SkillObjects import (
    Skill,
    SkillValue,
    SkillFunction,
    SkillFunctionArgument,
    MXClassproperty,
    MXValueType,
    MXType,
)

INFINITY = sys.maxsize


@dataclass
class airConditionerMode(Skill):

    @MXClassproperty
    def descriptor(cls) -> str:
        return 'Allows for the control of the air conditioner.'

    class Enums:
        @MXClassproperty
        def descriptor(cls) -> str:
            return textwrap.dedent(
                '''\
                • auto - The fan is on auto
                • cool - The fan is in sleep mode to reduce noise
                • heat - The fan is on low'''
            )

        class airConditionerModeEnum(MXStrEnum):
            auto = aauto()
            cool = aauto()
            heat = aauto()

    class Values:
        @dataclass
        class airConditionerMode(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Current mode of the air conditioner'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.ENUM, format=airConditionerMode.Enums.airConditionerModeEnum)

        @dataclass
        class temperatureSet(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Current temperature status of the air conditioner'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.INTEGER)

        class supportedAcModes(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Supported states for this air conditioner to be in'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.LIST, format=List[airConditionerMode.Enums.airConditionerModeEnum])

    class Functions:
        @dataclass
        class setAirConditionerMode(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Set the air conditioner mode'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.VOID

            class Arguments:
                @dataclass
                class mode(SkillFunctionArgument):
                    @MXClassproperty
                    def descriptor(cls) -> str:
                        return 'Set the air conditioner mode'

                    @MXClassproperty
                    def argument_type(cls) -> MXValueType:
                        return MXValueType(MXType.ENUM, format=airConditionerMode.Enums.airConditionerModeEnum)

        @dataclass
        class setTemperature(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Set the air conditioner temperature'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.VOID

            class Arguments:
                @dataclass
                class temperature(SkillFunctionArgument):
                    @MXClassproperty
                    def descriptor(cls) -> str:
                        return 'Set the air conditioner temperature'

                    @MXClassproperty
                    def argument_type(cls) -> MXValueType:
                        return MXValueType(MXType.INTEGER)


@dataclass
class airPurifierFanMode(Skill):

    @MXClassproperty
    def descriptor(cls) -> str:
        return 'Maintains and sets the state of an air purifier\'s fan'

    class Enums:
        @MXClassproperty
        def descriptor(cls) -> str:
            return textwrap.dedent(
                '''\
                • auto - The fan is on auto
                • sleep - The fan is in sleep mode to reduce noise
                • low - The fan is on low
                • medium - The fan is on medium
                • high - The fan is on high
                • quiet - The fan is on quiet mode to reduce noise
                • windFree - The fan is on wind free mode to reduce the feeling of cold air
                • off - The fan is off'''
            )

        class airPurifierFanModeEnum(MXStrEnum):
            auto = aauto()
            sleep = aauto()
            low = aauto()
            medium = aauto()
            high = aauto()
            quiet = aauto()
            windFree = aauto()
            off = aauto()

    class Values:
        @dataclass
        class airPurifierFanMode(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'The current mode of the air purifier fan, an enum of auto, low, medium, high, sleep, quiet or windFree'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.ENUM, format=airPurifierFanMode.Enums.airPurifierFanModeEnum)

        class supportedAirPurifierFanModes(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Supported states for this air purifier fan to be in'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.LIST, format=List[airPurifierFanMode.Enums.airPurifierFanModeEnum])

    class Functions:
        @dataclass
        class setAirPurifierFanMode(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Set the air purifier fan\'s mode'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.VOID

            class Arguments:
                @dataclass
                class mode(SkillFunctionArgument):
                    @MXClassproperty
                    def descriptor(cls) -> str:
                        return 'Set the air purifier fan\'s mode'

                    @MXClassproperty
                    def argument_type(cls) -> MXValueType:
                        return MXValueType(MXType.ENUM, format=airPurifierFanMode.Enums.airPurifierFanModeEnum)


@dataclass
class battery(Skill):

    @MXClassproperty
    def descriptor(cls) -> str:
        return 'Defines that the device has a battery'

    class Values:
        @dataclass
        class battery(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'An indication of the status of the battery'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.INTEGER, [0, 100])


@dataclass
class button(Skill):

    @MXClassproperty
    def descriptor(cls) -> str:
        return 'A device with one or more buttons'

    class Enums:
        @MXClassproperty
        def descriptor(cls) -> str:
            return textwrap.dedent(
                '''\
                • pushed - The value if the button is pushed
                • held - The value if the button is held
                • double - The value if the button is pushed twice
                • pushed_2x - The value if the button is pushed twice
                • pushed_3x - The value if the button is pushed three times
                • pushed_4x - The value if the button is pushed four times
                • pushed_5x - The value if the button is pushed five times
                • pushed_6x - The value if the button is pushed six times
                • down - The value if the button is clicked down
                • down_2x - The value if the button is clicked down twice
                • down_3x - The value if the button is clicked down three times
                • down_4x - The value if the button is clicked down four times
                • down_5x - The value if the button is clicked down five times
                • down_6x - The value if the button is clicked down six times
                • down_hold - The value if the button is clicked down and held
                • up - The value if the button is clicked up
                • up_2x - The value if the button is clicked up twice
                • up_3x - The value if the button is clicked up three times
                • up_4x - The value if the button is clicked up four times
                • up_5x - The value if the button is clicked up five times
                • up_6x - The value if the button is clicked up six times
                • up_hold - The value if the button is clicked up and held
                • swipe_up - The value if the button is swiped up from botton to top
                • swipe_down - The value if the button is swiped down from top to bottom
                • swipe_left - The value if the button is swiped from right to left
                • swipe_right - The value if the button is swiped from left to right'''
            )

        class buttonEnum(MXStrEnum):
            pushed = aauto()
            held = aauto()
            double = aauto()
            pushed_2x = aauto()
            pushed_3x = aauto()
            pushed_4x = aauto()
            pushed_5x = aauto()
            pushed_6x = aauto()
            down = aauto()
            down_2x = aauto()
            down_3x = aauto()
            down_4x = aauto()
            down_5x = aauto()
            down_6x = aauto()
            down_hold = aauto()
            up = aauto()
            up_2x = aauto()
            up_3x = aauto()
            up_4x = aauto()
            up_5x = aauto()
            up_6x = aauto()
            up_hold = aauto()
            swipe_up = aauto()
            swipe_down = aauto()
            swipe_left = aauto()
            swipe_right = aauto()

    class Values:
        @dataclass
        class button(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'The state of the buttons'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.ENUM, format=button.Enums.buttonEnum)

        @dataclass
        class numberOfButtons(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'The number of buttons on the device'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.INTEGER, bound=[0, INFINITY])

        @dataclass
        class supportedButtonValues(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'List of valid button attribute values'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.LIST, format=List[button.Enums.buttonEnum])


@dataclass
class carbonDioxideMeasurement(Skill):

    @MXClassproperty
    def descriptor(cls) -> str:
        return 'Measure carbon dioxide levels'

    class Values:
        @dataclass
        class carbonDioxide(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'The level of carbon dioxide detected'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.DOUBLE, [0, 1000000])


@dataclass
class chargingState(Skill):

    @MXClassproperty
    def descriptor(cls) -> str:
        return 'The current status of battery charging'

    class Enums:
        @MXClassproperty
        def descriptor(cls) -> str:
            return textwrap.dedent(
                '''\
                • charging - charging
                • discharging - discharging
                • stopped - stopped
                • fullyCharged - fully charged
                • error - error'''
            )

        class chargingStateEnum(MXStrEnum):
            charging = aauto()
            discharging = aauto()
            stopped = aauto()
            fullyCharged = aauto()
            error = aauto()

    class Values:
        @dataclass
        class chargingState(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'The current charging state of the device'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.ENUM, format=chargingState.Enums.chargingStateEnum)

        @dataclass
        class supportedChargingStates(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'The list of charging states that the device supports. Optional, defaults to all states if not set.'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.LIST, format=List[chargingState.Enums.chargingStateEnum])


@dataclass
class clock(Skill):

    @MXClassproperty
    def descriptor(cls) -> str:
        return 'Provide current date and time'

    class Enums:
        @MXClassproperty
        def descriptor(cls) -> str:
            return textwrap.dedent(
                '''\
                • monday
                • tuesday
                • wednesday
                • thursday
                • friday
                • saturday
                • sunday'''
            )

        class weekdayEnum(MXStrEnum):
            monday = aauto()
            tuesday = aauto()
            wednesday = aauto()
            thursday = aauto()
            friday = aauto()
            saturday = aauto()
            sunday = aauto()

    class Values:
        @dataclass
        class year(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Current year'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.INTEGER, [0, 100000])

        @dataclass
        class month(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Current month'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.INTEGER, [1, 12])

        @dataclass
        class day(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Current day'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.INTEGER, [1, 31])

        @dataclass
        class weekday(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Current weekday'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.ENUM, format=clock.Enums.weekdayEnum)

        @dataclass
        class hour(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Current hour'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.INTEGER, [0, 24])

        @dataclass
        class minute(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Current minute'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.INTEGER, [0, 60])

        @dataclass
        class second(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Current second'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.INTEGER, [0, 60])

        @dataclass
        class timestamp(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Current timestamp (return current unix time - unit: seconds with floating point)'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.DOUBLE)

        @dataclass
        class datetime(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Current date and time as double number - format: YYYYMMddhhmm'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.DOUBLE)

        @dataclass
        class date(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Current date as double number - format: YYYYMMdd'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.DOUBLE)

        @dataclass
        class time(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Current time as double number - format: hhmm'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.DOUBLE)

        @dataclass
        class isHoliday(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'today is holiday or not'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.BOOL)


@dataclass
class contactSensor(Skill):

    @MXClassproperty
    def descriptor(cls) -> str:
        return 'Allows reading the value of a contact sensor device'

    class Enums:
        @MXClassproperty
        def descriptor(cls) -> str:
            return textwrap.dedent(
                '''\
                • closed - The value if closed
                • open - The value if open'''
            )

        class contactEnum(MXStrEnum):
            open = aauto()
            closed = aauto()

    class Values:
        @dataclass
        class contact(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'The current state of the contact sensor'

            @MXClassproperty
            def value_type(cls) -> MXType:
                return MXValueType(MXType.ENUM, format=contactSensor.Enums.contactEnum)


@dataclass
class colorControl(Skill):

    @MXClassproperty
    def descriptor(cls) -> str:
        return 'Allows for control of a color changing device by setting its hue, saturation, and color values'

    class Values:
        @dataclass
        class color(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return '``{"hue":"0-100 (percent)", "saturation":"0-100 (percent)"}``'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.STRING)

        @dataclass
        class hue(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return '``0-100`` (percent)'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.DOUBLE, [0, 100])

        @dataclass
        class saturation(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return '``0-100`` (percent)'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.DOUBLE, [0, 100])

    class Functions:
        @dataclass
        class setColor(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Sets the color based on the values passed in with the given map'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.VOID

            class Arguments:
                @dataclass
                class color(SkillFunctionArgument):
                    @MXClassproperty
                    def descriptor(cls) -> str:
                        return 'The color map supports the following key/value pairs:'

                    @MXClassproperty
                    def argument_type(cls) -> MXValueType:
                        return MXValueType(MXType.DICT)

        @dataclass
        class setHue(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Set the hue value of the color'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.VOID

            class Arguments:
                @dataclass
                class hue(SkillFunctionArgument):
                    @MXClassproperty
                    def descriptor(cls) -> str:
                        return 'A number in the range ``0-100`` representing the hue as a value of percent'

                    @MXClassproperty
                    def argument_type(cls) -> MXValueType:
                        return MXValueType(MXType.DOUBLE, [0, 100])

        @dataclass
        class setSaturation(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Set the saturation value of the color'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.VOID

            class Arguments:
                @dataclass
                class saturation(SkillFunctionArgument):
                    @MXClassproperty
                    def descriptor(cls) -> str:
                        return 'A number in the range ``0-100`` representing the saturation as a value of percent'

                    @MXClassproperty
                    def argument_type(cls) -> MXValueType:
                        return MXValueType(MXType.DOUBLE, [0, 100])


@dataclass
class refresh(Skill):

    @MXClassproperty
    def descriptor(cls) -> str:
        return 'Allow the execution of the refresh command for devices that support it'

    class Functions:
        @dataclass
        class refresh(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Send the refresh command to the device - Return refresh success'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.BOOL


@dataclass
class currentMeasurement(Skill):

    @MXClassproperty
    def descriptor(cls) -> str:
        return 'Get the value of electrical current measured from a device.'

    class Values:
        @dataclass
        class current(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'A number representing the current measured.'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.DOUBLE)


@dataclass
class dehumidifierMode(Skill):

    @MXClassproperty
    def descriptor(cls) -> str:
        return 'Allows for the control of the dehumidifier mode.'

    class Enums:
        @MXClassproperty
        def descriptor(cls) -> str:
            return textwrap.dedent(
                '''\
                • cooling
                • delayWash
                • drying
                • finished
                • refreshing
                • weightSensing
                • wrinklePrevent
                • dehumidifying
                • AIDrying
                • sanitizing
                • internalCare
                • freezeProtection
                • continuousDehumidifying
                • thawingFrozenInside'''
            )

        class dehumidifierModeEnum(MXStrEnum):
            cooling = aauto()
            delayWash = aauto()
            drying = aauto()
            finished = aauto()
            refreshing = aauto()
            weightSensing = aauto()
            wrinklePrevent = aauto()
            dehumidifying = aauto()
            AIDrying = aauto()
            sanitizing = aauto()
            internalCare = aauto()
            freezeProtection = aauto()
            continuousDehumidifying = aauto()
            thawingFrozenInside = aauto()

    class Values:
        @dataclass
        class dehumidifierMode(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Current mode of the dehumidifier'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.ENUM, format=dehumidifierMode.Enums.dehumidifierModeEnum)

    class Functions:
        @dataclass
        class setDehumidifierMode(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Set the dehumidifier mode'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.VOID

            class Arguments:
                @dataclass
                class mode(SkillFunctionArgument):
                    @MXClassproperty
                    def descriptor(cls) -> str:
                        return 'Set the dehumidifier mode'

                    @MXClassproperty
                    def argument_type(cls) -> MXValueType:
                        return MXValueType(MXType.ENUM, format=dehumidifierMode.Enums.dehumidifierModeEnum)


@dataclass
class dishwasherMode(Skill):

    @MXClassproperty
    def descriptor(cls) -> str:
        return 'Allows for the control of the dishwasher mode.'

    class Enums:
        @MXClassproperty
        def descriptor(cls) -> str:
            return textwrap.dedent(
                '''\
                • eco - The dishwasher is in "eco" mode
                • intense - The dishwasher is in "intense" mode
                • auto - The dishwasher is in "auto" mode
                • quick - The dishwasher is in "quick" mode
                • rinse - The dishwasher is in "rinse" mode
                • dry - The dishwasher is in "dry" mode'''
            )

        class dishwasherModeEnum(MXStrEnum):
            eco = aauto()
            intense = aauto()
            auto = aauto()
            quick = aauto()
            rinse = aauto()
            dry = aauto()

    class Values:
        @dataclass
        class dishwasherMode(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Current mode of the dishwasher'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.ENUM, format=dishwasherMode.Enums.dishwasherModeEnum)

    class Functions:
        @dataclass
        class setDishwasherMode(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Set the dishwasher mode'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.VOID

            class Arguments:
                @dataclass
                class mode(SkillFunctionArgument):
                    @MXClassproperty
                    def descriptor(cls) -> str:
                        return 'Set the dishwasher mode to "eco", "intense", "auto", "quick", "rinse", or "dry" mode'

                    @MXClassproperty
                    def argument_type(cls) -> MXValueType:
                        return MXValueType(MXType.ENUM, format=dishwasherMode.Enums.dishwasherModeEnum)


@dataclass
class doorControl(Skill):

    @MXClassproperty
    def descriptor(cls) -> str:
        return 'Allow for the control of a door'

    class Enums:
        @MXClassproperty
        def descriptor(cls) -> str:
            return textwrap.dedent(
                '''\
                • closed - The door is closed
                • closing - The door is closing
                • open - The door is open
                • opening - The door is opening
                • unknown - The current state of the door is unknown'''
            )

        class doorEnum(MXStrEnum):
            closed = aauto()
            closing = aauto()
            open = aauto()
            opening = aauto()
            unknown = aauto()

    class Values:
        @dataclass
        class door(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'The current state of the door'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.ENUM, format=doorControl.Enums.doorEnum)

    class Functions:
        @dataclass
        class open(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Open the door'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.VOID

        @dataclass
        class close(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Close the door'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.VOID


@dataclass
class dustSensor(Skill):

    @MXClassproperty
    def descriptor(cls) -> str:
        return 'Gets the reading of the dust sensor.'

    class Values:
        @dataclass
        class dustLevel(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Current dust level -- also refered to as PM10, measured in micrograms per cubic meter'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.INTEGER, [0, INFINITY])

        @dataclass
        class fineDustLevel(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Current level of fine dust -- also refered to as PM2.5, measured in micrograms per cubic meter'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.INTEGER, [0, INFINITY])

        @dataclass
        class veryFineDustLevel(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Current level of fine dust -- also refered to as PM1.0, measured in micrograms per cubic meter'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.INTEGER, [0, INFINITY])


@dataclass
class fanControl(Skill):

    @MXClassproperty
    def descriptor(cls) -> str:
        return 'Allows for the control of the fan.'

    class Values:
        @dataclass
        class fanSpeed(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'The current fan speed represented as a integer value. - unit: RPM'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.INTEGER, [0, INFINITY])

        @dataclass
        class percent(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'The current fan speed represented as a percent value.'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.INTEGER, [0, 100])

    class Functions:
        @dataclass
        class setFanSpeed(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Set the fan speed'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.VOID

            class Arguments:
                @dataclass
                class speed(SkillFunctionArgument):
                    @MXClassproperty
                    def descriptor(cls) -> str:
                        return 'Set the fan to this speed'

                    @MXClassproperty
                    def argument_type(cls) -> MXValueType:
                        return MXValueType(MXType.INTEGER, [0, INFINITY])

        @dataclass
        class setPercent(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Set the fan speed percent.'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.VOID

            class Arguments:
                @dataclass
                class percent(SkillFunctionArgument):
                    @MXClassproperty
                    def descriptor(cls) -> str:
                        return 'The percent value to set the fan speed to.'

                    @MXClassproperty
                    def argument_type(cls) -> MXValueType:
                        return MXValueType(MXType.INTEGER, [0, 100])


@dataclass
class feederOperatingState(Skill):

    @MXClassproperty
    def descriptor(cls) -> str:
        return 'Allows for the control of a feeder device.'

    class Enums:
        @MXClassproperty
        def descriptor(cls) -> str:
            return textwrap.dedent(
                '''\
                • idle - idle
                • feeding - feeding
                • error - error'''
            )

        class feederOperatingStateEnum(MXStrEnum):
            idle = aauto()
            feeding = aauto()
            error = aauto()

    class Values:
        @dataclass
        class feederOperatingState(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'The current state of the feeder.'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.ENUM, format=feederOperatingState.Enums.feederOperatingStateEnum)

    class Functions:
        @dataclass
        class startFeeding(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Begin the feeding process.'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.VOID


@dataclass
class irrigatorOperatingState(Skill):

    @MXClassproperty
    def descriptor(cls) -> str:
        return 'Allows for the control of a irrigator device.'

    class Enums:
        @MXClassproperty
        def descriptor(cls) -> str:
            return textwrap.dedent(
                '''\
                • idle - idle
                • watering - watering
                • error - error'''
            )

        class irrigatorOperatingStateEnum(MXStrEnum):
            idle = aauto()
            watering = aauto()
            error = aauto()

    class Values:
        @dataclass
        class irrigatorOperatingState(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'The current state of the irrigator.'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.ENUM, format=irrigatorOperatingState.Enums.irrigatorOperatingStateEnum)

    class Functions:
        @dataclass
        class startWatering(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Begin the watering process.'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.VOID


@dataclass
class feederPortion(Skill):

    @MXClassproperty
    def descriptor(cls) -> str:
        return 'Allows for the portion control of a feeder device.'

    class Enums:
        @MXClassproperty
        def descriptor(cls) -> str:
            return textwrap.dedent(
                '''\
                • grams
                • pounds
                • ounces
                • servings'''
            )

        class unitEnum(MXStrEnum):
            grams = aauto()
            pounds = aauto()
            ounces = aauto()
            servings = aauto()

    class Values:
        @dataclass
        class feedPortion(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'A number that represents the portion (in grams, pounds, ounces, or servings) that will dispense.'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.DOUBLE, [0, 2000])

    class Functions:
        @dataclass
        class setFeedPortion(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Set the portion (in grams, pounds, ounces, or servings) that the feeder will dispense.'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.VOID

            class Arguments:
                @dataclass
                class portion(SkillFunctionArgument):
                    @MXClassproperty
                    def descriptor(cls) -> str:
                        return 'The portion (in grams, pounds, ounces, or servings) to dispense.'

                    @MXClassproperty
                    def argument_type(cls) -> MXValueType:
                        return MXValueType(MXType.DOUBLE, [0, 2000])

                @dataclass
                class unit(SkillFunctionArgument):
                    @MXClassproperty
                    def descriptor(cls) -> str:
                        return ''

                    @MXClassproperty
                    def argument_type(cls) -> MXValueType:
                        return MXValueType(MXType.ENUM, format=feederPortion.Enums.unitEnum)


@dataclass
class irrigatorPortion(Skill):

    @MXClassproperty
    def descriptor(cls) -> str:
        return 'Allows for the portion control of a irrigator device.'

    class Enums:
        @MXClassproperty
        def descriptor(cls) -> str:
            return textwrap.dedent(
                '''\
                • liters
                • milliliters
                • gallons
                • ounces
                '''
            )

        class unitEnum(MXStrEnum):
            liters = aauto()
            milliliters = aauto()
            gallons = aauto()
            ounces = aauto()

    class Values:
        @dataclass
        class waterPortion(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'A number that represents the portion (in liters, milliliters, gallons, or ounces) that will dispense.'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.DOUBLE)

    class Functions:
        @dataclass
        class setWaterPortion(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Set the portion (in liters, milliliters, gallons, or ounces) that the irrigator will dispense.'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.VOID

            class Arguments:
                @dataclass
                class portion(SkillFunctionArgument):
                    @MXClassproperty
                    def descriptor(cls) -> str:
                        return 'The portion (in grams, pounds, ounces, or servings) to dispense.'

                    @MXClassproperty
                    def argument_type(cls) -> MXValueType:
                        return MXValueType(MXType.DOUBLE, [0, 2000])

                @dataclass
                class unit(SkillFunctionArgument):
                    @MXClassproperty
                    def descriptor(cls) -> str:
                        return ''

                    @MXClassproperty
                    def argument_type(cls) -> MXValueType:
                        return MXValueType(MXType.ENUM, format=irrigatorPortion.Enums.unitEnum)


@dataclass
class switch(Skill):

    @MXClassproperty
    def descriptor(cls) -> str:
        return 'Allows for the control of a switch device'

    class Enums:
        @MXClassproperty
        def descriptor(cls) -> str:
            return textwrap.dedent(
                '''\
                • on - The value of the ``switch`` attribute if the switch is on
                • off - The value of the ``switch`` attribute if the switch is off'''
            )

        class switchEnum(MXStrEnum):
            on = aauto()
            off = aauto()

    class Values:
        @dataclass
        class switch(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'A string representation of whether the switch is on or off'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.ENUM, format=switch.Enums.switchEnum)

    class Functions:
        @dataclass
        class on(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Turn a switch on'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.VOID

        @dataclass
        class off(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Turn a switch off'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.VOID


@dataclass
class switchLevel(Skill):

    @MXClassproperty
    def descriptor(cls) -> str:
        return 'Allows for the control of the level of a device like a light or a dimmer switch.'

    class Values:
        @dataclass
        class level(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'A number that represents the current level, usually ``0-100`` in percent'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.INTEGER, [0, 100])

        @dataclass
        class levelRange(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Constraints on the level value'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.DICT)

    class Functions:
        @dataclass
        class setLevel(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Set the level to the given value. If the device supports being turned on and off then it will be turned on if ``level`` is greater than 0 and turned off if ``level`` is equal to 0.'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.VOID

            class Arguments:
                @dataclass
                class level(SkillFunctionArgument):
                    @MXClassproperty
                    def descriptor(cls) -> str:
                        return 'The level value, usually ``0-100`` in percent'

                    @MXClassproperty
                    def argument_type(cls) -> MXValueType:
                        return MXValueType(MXType.INTEGER, [-0, 100])

                @dataclass
                class rate(SkillFunctionArgument):
                    @MXClassproperty
                    def descriptor(cls) -> str:
                        return 'The rate at which to change the level'

                    @MXClassproperty
                    def argument_type(cls) -> MXValueType:
                        return MXValueType(MXType.INTEGER, [-0, 100])

        @dataclass
        class alert(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Alert with dimming'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.VOID


@dataclass
class lightLevel(Skill):

    @MXClassproperty
    def descriptor(cls) -> str:
        return 'A numerical representation of the brightness intensity'

    class Values:
        @dataclass
        class light(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'brightness intensity (Unit: lux)'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.DOUBLE, [0, INFINITY])


@dataclass
class manager(Skill):

    @MXClassproperty
    def descriptor(cls) -> str:
        return 'Allow Manager Thing\'s features'

    class Functions:
        @dataclass
        class discover(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Discover local devices - Return device list with json format'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.STRING

        @dataclass
        class add_thing(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Add staff thing - Return success or fail'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.BOOL

            class Arguments:
                @dataclass
                class staff_thing_info(SkillFunctionArgument):
                    @MXClassproperty
                    def descriptor(cls) -> str:
                        return 'Staff Thing\'s information'

                    @MXClassproperty
                    def argument_type(cls) -> MXValueType:
                        return MXValueType(MXType.STRING)

                @dataclass
                class client_id(SkillFunctionArgument):
                    @MXClassproperty
                    def descriptor(cls) -> str:
                        return 'Requester\'s client id'

                    @MXClassproperty
                    def argument_type(cls) -> MXValueType:
                        return MXValueType(MXType.STRING)

        @dataclass
        class delete_thing(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Delete staff thing - Return success or fail'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.BOOL

            class Arguments:
                @dataclass
                class staff_thing_info(SkillFunctionArgument):
                    @MXClassproperty
                    def descriptor(cls) -> str:
                        return 'Staff Thing\'s information'

                    @MXClassproperty
                    def argument_type(cls) -> MXValueType:
                        return MXValueType(MXType.STRING)

                @dataclass
                class client_id(SkillFunctionArgument):
                    @MXClassproperty
                    def descriptor(cls) -> str:
                        return 'Requester\'s client id'

                    @MXClassproperty
                    def argument_type(cls) -> MXValueType:
                        return MXValueType(MXType.STRING)


@dataclass
class motionSensor(Skill):

    @MXClassproperty
    def descriptor(cls) -> str:
        return textwrap.dedent(
            '''\
                • active - The value when motion is detected
                • inactive - The value when no motion is detected'''
        )

    class Enums:
        @MXClassproperty
        def descriptor(cls) -> str:
            return textwrap.dedent('''  ''')

        class motionEnum(MXStrEnum):
            active = aauto()
            inactive = aauto()

    class Values:
        @dataclass
        class motion(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'The current state of the motion sensor'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.ENUM, format=motionSensor.Enums.motionEnum)


@dataclass
class temperatureMeasurement(Skill):

    @MXClassproperty
    def descriptor(cls) -> str:
        return 'Get the temperature from a Device that reports current temperature'

    class Values:
        @dataclass
        class temperature(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'A number that usually represents the current temperature'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.INTEGER, [-460, 10000])

        @dataclass
        class temperatureRange(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Constraints on the temperature value'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.DICT)


@dataclass
class testSkill(Skill):

    @MXClassproperty
    def descriptor(cls) -> str:
        return 'testSkill'

    class Enums:
        @MXClassproperty
        def descriptor(cls) -> str:
            return textwrap.dedent('''testSkill Enums''')

        class testSkillEnum(MXStrEnum):
            enum1 = aauto()
            enum2 = aauto()
            enum3 = aauto()

    class Values:
        @dataclass
        class testSkillValue(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'testSkillValue'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.STRING)

    class Functions:
        @dataclass
        class testSkillFunction(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'testSkillFunction'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.STRING

            class Arguments:
                @dataclass
                class testArgument(SkillFunctionArgument):
                    @MXClassproperty
                    def descriptor(cls) -> str:
                        return 'testArgument'

                    @MXClassproperty
                    def argument_type(cls) -> MXValueType:
                        return MXValueType(MXType.STRING)


@dataclass
class tvChannel(Skill):

    @MXClassproperty
    def descriptor(cls) -> str:
        return 'Allows for the control of the TV channel.'

    class Values:
        @dataclass
        class tvChannel(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Current status of the TV channel'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.INTEGER)

        @dataclass
        class tvChannelName(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Current status of the TV channel name'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.STRING)

    class Functions:
        @dataclass
        class channelUp(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Move the TV channel up'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.VOID

        @dataclass
        class channelDown(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Move the TV channel down'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.VOID

        @dataclass
        class setTvChannel(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Set the TV channel'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.VOID

            class Arguments:
                @dataclass
                class tvChannel(SkillFunctionArgument):
                    @MXClassproperty
                    def descriptor(cls) -> str:
                        return ''

                    @MXClassproperty
                    def argument_type(cls) -> MXValueType:
                        return MXValueType(MXType.INTEGER)

        @dataclass
        class setTvChannelName(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Set the TV channel Name'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.VOID

            class Arguments:
                @dataclass
                class tvChannelName(SkillFunctionArgument):
                    @MXClassproperty
                    def descriptor(cls) -> str:
                        return ''

                    @MXClassproperty
                    def argument_type(cls) -> MXValueType:
                        return MXValueType(MXType.STRING)


@dataclass
class voltageMeasurement(Skill):

    @MXClassproperty
    def descriptor(cls) -> str:
        return 'Get the value of voltage measured from devices that support it'

    class Values:
        @dataclass
        class voltage(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'A number representing the current voltage measured'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.DOUBLE)


@dataclass
class gasMeter(Skill):

    @MXClassproperty
    def descriptor(cls) -> str:
        return 'Read the gas consumption of an energy metering device'

    class Values:
        @dataclass
        class gasMeter(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'the gas energy reported by the metering device. unit: kWh'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.DOUBLE)

        @dataclass
        class gasMeterCalorific(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'a measure of the available heat energy, used as part of the calculation to convert gas volume to gas energy. - unit: kcal'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.DOUBLE)

        @dataclass
        class gasMeterTime(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'The cumulative gas use time reported by the metering device. - unit: seconds'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.DOUBLE)

        @dataclass
        class gasMeterVolume(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'the cumulative gas volume reported by the metering device. - unit: cubic meters'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.DOUBLE)


@dataclass
class soilHumidityMeasurement(Skill):

    @MXClassproperty
    def descriptor(cls) -> str:
        return 'Allow reading the soil humidity from devices that support it'

    class Values:
        @dataclass
        class soilHumidity(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'A numerical representation of the soil humidity measurement taken by the device'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.DOUBLE)


@dataclass
class camera(Skill):

    @MXClassproperty
    def descriptor(cls) -> str:
        return 'Allows for the control of a camera device'

    class Values:
        @dataclass
        class image(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'The latest image captured by the camera'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.BINARY)

        @dataclass
        class video(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'The latest video captured by the camera'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.BINARY)

    class Functions:
        @dataclass
        class take(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Take a picture with the camera - Return the image as binary data'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.BINARY

        @dataclass
        class takeTimelapse(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Take a picture with the camera - Return the video as binary data'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.BINARY

            class Arguments:

                @dataclass
                class duration(SkillFunctionArgument):
                    @MXClassproperty
                    def descriptor(cls) -> str:
                        return 'The duration of the timelapse in seconds'

                    @MXClassproperty
                    def argument_type(cls) -> MXValueType:
                        return MXValueType(MXType.DOUBLE, [0, INFINITY])

                @dataclass
                class speed(SkillFunctionArgument):
                    @MXClassproperty
                    def descriptor(cls) -> str:
                        return 'The speed of the timelapse'

                    @MXClassproperty
                    def argument_type(cls) -> MXValueType:
                        return MXValueType(MXType.DOUBLE, [0, INFINITY])


@dataclass
class tvocMeasurement(Skill):

    @MXClassproperty
    def descriptor(cls) -> str:
        return 'Measure total volatile organic compound levels'

    class Values:
        @dataclass
        class tvocLevel(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'The level of total volatile organic compounds detected'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.DOUBLE, [0, 1000000])


@dataclass
class valve(Skill):

    @MXClassproperty
    def descriptor(cls) -> str:
        return 'Allows for the control of a valve device'

    class Enums:
        @MXClassproperty
        def descriptor(cls) -> str:
            return textwrap.dedent(
                '''\
                • closed - The value of the ``valve`` attribute if the valve is closed
                • open - The value of the ``valve`` attribute if the valve is open'''
            )

        class valveEnum(MXStrEnum):
            open = aauto()
            closed = aauto()

    class Values:
        @dataclass
        class valve(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'A string representation of whether the valve is open or closed'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.ENUM, format=valve.Enums.valveEnum)

    class Functions:
        @dataclass
        class open(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Open the valve'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.VOID

        @dataclass
        class close(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Close the valve'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.VOID


@dataclass
class pump(Skill):

    @MXClassproperty
    def descriptor(cls) -> str:
        return 'Allows for the control of a pump device'

    class Enums:
        @MXClassproperty
        def descriptor(cls) -> str:
            return textwrap.dedent(
                '''\
                • closed - The value of the ``pump`` attribute if the pump is closed
                • open - The value of the ``pump`` attribute if the pump is open'''
            )

        class pumpEnum(MXStrEnum):
            open = aauto()
            closed = aauto()

    class Values:
        @dataclass
        class pump(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'A string representation of whether the pump is open or closed'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.ENUM, format=pump.Enums.pumpEnum)

    class Functions:
        @dataclass
        class open(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Open the pump'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.VOID

        @dataclass
        class close(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Close the pump'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.VOID


@dataclass
class audioMute(Skill):

    @MXClassproperty
    def descriptor(cls) -> str:
        return 'Allows for the control of audio mute.'

    class Enums:
        @MXClassproperty
        def descriptor(cls) -> str:
            return textwrap.dedent(
                '''\
                • muted - The audio is in "muted" state
                • unmuted - The audio is in "unmuted" state'''
            )

        class muteEnum(MXStrEnum):
            muted = aauto()
            unmuted = aauto()

    class Values:
        @dataclass
        class muteStatus(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Current status of the audio mute'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.ENUM, format=audioMute.Enums.muteEnum)

    class Functions:
        @dataclass
        class mute(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Set the audio to mute state'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.VOID

        @dataclass
        class unmute(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Set the audio to unmute state'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.VOID

        @dataclass
        class setMute(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Set the state of the audio mute'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.VOID

            class Arguments:
                @dataclass
                class state(SkillFunctionArgument):
                    @MXClassproperty
                    def descriptor(cls) -> str:
                        return 'Set the audio mute state to "muted" or "unmuted"'

                    @MXClassproperty
                    def argument_type(cls) -> MXValueType:
                        return MXValueType(MXType.ENUM, format=audioMute.Enums.muteEnum)


@dataclass
class mediaPlayback(Skill):

    @MXClassproperty
    def descriptor(cls) -> str:
        return 'Allows for the control of the media playback.'

    class Enums:
        @MXClassproperty
        def descriptor(cls) -> str:
            return textwrap.dedent(
                '''\
                • paused - Media playback is in a "paused" state
                • playing - Media playback is in a "playing" state
                • stopped - Media playback is in a "stopped" state
                • fast forwarding - Media playback is in a "fast forwarding" state
                • rewinding - Media playback is in a "rewinding" state
                • buffering - Media playback is in a "buffering" state'''
            )

        class mediaPlaybackEnum(MXStrEnum):
            paused = aauto()
            playing = aauto()
            stopped = aauto()
            fast = aauto()
            rewinding = aauto()
            buffering = aauto()

    class Values:
        @dataclass
        class playbackStatus(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Status of the media playback'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.ENUM, format=mediaPlayback.Enums.mediaPlaybackEnum)

        @dataclass
        class supportedPlaybackCommands(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Media playback commands which are supported'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.LIST, format=List[mediaPlayback.Enums.mediaPlaybackEnum])

    class Functions:

        @dataclass
        class play(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Play the media playback'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.VOID

            class Arguments:
                @dataclass
                class source(SkillFunctionArgument):
                    @MXClassproperty
                    def descriptor(cls) -> str:
                        return 'The source of the media playback'

                    @MXClassproperty
                    def argument_type(cls) -> MXValueType:
                        return MXValueType(MXType.STRING)

        @dataclass
        class stop(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Stop the media playback'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.VOID

        @dataclass
        class pause(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Pause the media playback'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.VOID

        @dataclass
        class fastForward(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Fast forward the media playback'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.VOID

        @dataclass
        class rewind(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Rewind the media playback'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.VOID

        @dataclass
        class setPlaybackStatus(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Set the playback status'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.VOID

            class Arguments:
                @dataclass
                class status(SkillFunctionArgument):
                    @MXClassproperty
                    def descriptor(cls) -> str:
                        return 'Set the playback status to "paused", "playing", "stopped", "fast forwarding" or "rewinding" state.'

                    @MXClassproperty
                    def argument_type(cls) -> MXValueType:
                        return MXValueType(MXType.ENUM, format=mediaPlayback.Enums.mediaPlaybackEnum)

        @dataclass
        class speak(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'TTS feature'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.VOID

            class Arguments:
                @dataclass
                class text(SkillFunctionArgument):
                    @MXClassproperty
                    def descriptor(cls) -> str:
                        return 'The text to be spoken'

                    @MXClassproperty
                    def argument_type(cls) -> MXValueType:
                        return MXValueType(MXType.STRING)


@dataclass
class refrigeration(Skill):

    @MXClassproperty
    def descriptor(cls) -> str:
        return 'Allows for the control of the refrigeration.'

    class Enums:
        @MXClassproperty
        def descriptor(cls) -> str:
            return textwrap.dedent(
                '''\
                • on - The value of the ``defrost``, ``rapidCooling``, ``rapidFreezing`` attribute if the defrost, rapidCooling, rapidFreezing is on
                • off - The value of the ``defrost``, ``rapidCooling``, ``rapidFreezing`` attribute if the defrost, rapidCooling, rapidFreezing is off'''
            )

        class defrostEnum(MXStrEnum):
            on = aauto()
            off = aauto()

        class rapidCoolingEnum(MXStrEnum):
            on = aauto()
            off = aauto()

        class rapidFreezingEnum(MXStrEnum):
            on = aauto()
            off = aauto()

    class Values:
        @dataclass
        class defrost(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Status of the defrost'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.ENUM, format=refrigeration.Enums.defrostEnum)

        @dataclass
        class rapidCooling(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Status of the rapid cooling'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.ENUM, format=refrigeration.Enums.rapidCoolingEnum)

        @dataclass
        class rapidFreezing(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Status of the rapid freezing'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.ENUM, format=refrigeration.Enums.rapidFreezingEnum)

    class Functions:
        @dataclass
        class setDefrost(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Sets the defrost on or off'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.VOID

            class Arguments:
                @dataclass
                class defrost(SkillFunctionArgument):
                    @MXClassproperty
                    def descriptor(cls) -> str:
                        return 'The on or off value for the defrost'

                    @MXClassproperty
                    def argument_type(cls) -> MXValueType:
                        return MXValueType(MXType.ENUM, format=refrigeration.Enums.defrostEnum)

        @dataclass
        class setRapidCooling(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Sets the rapid cooling on or off'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.VOID

            class Arguments:
                @dataclass
                class rapidCooling(SkillFunctionArgument):
                    @MXClassproperty
                    def descriptor(cls) -> str:
                        return 'The on or off value for the rapid cooling'

                    @MXClassproperty
                    def argument_type(cls) -> MXValueType:
                        return MXValueType(MXType.ENUM, format=refrigeration.Enums.rapidCoolingEnum)

        @dataclass
        class setRapidFreezing(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Sets the rapid freezing on or off'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.VOID

            class Arguments:
                @dataclass
                class rapidFreezing(SkillFunctionArgument):
                    @MXClassproperty
                    def descriptor(cls) -> str:
                        return 'The on or off value for the rapid freezing'

                    @MXClassproperty
                    def argument_type(cls) -> MXValueType:
                        return MXValueType(MXType.ENUM, format=refrigeration.Enums.rapidFreezingEnum)


@dataclass
class audioVolume(Skill):

    @MXClassproperty
    def descriptor(cls) -> str:
        return 'Allows for the control of audio volume.'

    class Values:
        @dataclass
        class volume(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'The current volume setting of the audio'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.INTEGER, [0, 100])

    class Functions:
        @dataclass
        class setVolume(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Set the audio volume level'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.VOID

            class Arguments:
                @dataclass
                class volume(SkillFunctionArgument):
                    @MXClassproperty
                    def descriptor(cls) -> str:
                        return 'A value to which the audio volume level should be set'

                    @MXClassproperty
                    def argument_type(cls) -> MXValueType:
                        return MXValueType(MXType.INTEGER, [0, 100])

        @dataclass
        class volumeUp(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Turn the audio volume up'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.VOID

        @dataclass
        class volumeDown(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Turn the audio volume down'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.VOID


@dataclass
class windowShadeLevel(Skill):

    @MXClassproperty
    def descriptor(cls) -> str:
        return 'Allows for the control of the level of a window shade.'

    class Values:
        @dataclass
        class shadeLevel(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'A number that represents the current level as a function of being open, ``0-100`` in percent; 0 representing completely closed, and 100 representing completely open.'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.INTEGER, [0, 100])

    class Functions:
        @dataclass
        class setShadeLevel(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Set the shade level to the given value.'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.VOID

            class Arguments:
                @dataclass
                class shadeLevel(SkillFunctionArgument):
                    @MXClassproperty
                    def descriptor(cls) -> str:
                        return 'The level to which the shade should be set, ``0-100`` in percent; 0 representing completely closed, and 100 representing completely open.'

                    @MXClassproperty
                    def argument_type(cls) -> MXValueType:
                        return MXValueType(MXType.INTEGER, [0, 100])


@dataclass
class curtainLevel(Skill):

    @MXClassproperty
    def descriptor(cls) -> str:
        return 'Allows for the control of the level of a curtain.'

    class Values:
        @dataclass
        class curtainLevel(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'A number that represents the current level as a function of being open, ``0-100`` in percent; 0 representing completely closed, and 100 representing completely open.'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.INTEGER, [0, 100])

    class Functions:
        @dataclass
        class setCurtainLevel(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Set the curtain level to the given value.'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.VOID

            class Arguments:
                @dataclass
                class curtainLevel(SkillFunctionArgument):
                    @MXClassproperty
                    def descriptor(cls) -> str:
                        return 'The level to which the curtain should be set, ``0-100`` in percent; 0 representing completely closed, and 100 representing completely open.'

                    @MXClassproperty
                    def argument_type(cls) -> MXValueType:
                        return MXValueType(MXType.INTEGER, [0, 100])


@dataclass
class blindLevel(Skill):

    @MXClassproperty
    def descriptor(cls) -> str:
        return 'Allows for the control of the level of a blind.'

    class Values:
        @dataclass
        class blindLevel(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'A number that represents the current level as a function of being open, ``0-100`` in percent; 0 representing completely closed, and 100 representing completely open.'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.INTEGER, [0, 100])

    class Functions:
        @dataclass
        class setBlindLevel(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Set the blind level to the given value.'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.VOID

            class Arguments:
                @dataclass
                class blindLevel(SkillFunctionArgument):
                    @MXClassproperty
                    def descriptor(cls) -> str:
                        return 'The level to which the blind should be set, ``0-100`` in percent; 0 representing completely closed, and 100 representing completely open.'

                    @MXClassproperty
                    def argument_type(cls) -> MXValueType:
                        return MXValueType(MXType.INTEGER, [0, 100])


@dataclass
class windowShade(Skill):

    @MXClassproperty
    def descriptor(cls) -> str:
        return 'Allows for the control of the window shade.'

    class Enums:
        @MXClassproperty
        def descriptor(cls) -> str:
            return textwrap.dedent(
                '''\
                • closed - closed
                • closing - closing…
                • open - open
                • opening - opening…
                • partially open - partially open
                • paused -
                • unknown - unknown'''
            )

        class windowShadeEnum(MXStrEnum):
            closed = aauto()
            closing = aauto()
            open = aauto()
            opening = aauto()
            partially = aauto()
            paused = aauto()
            unknown = aauto()

    class Values:
        @dataclass
        class windowShade(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'A string representation of whether the window shade is open or closed'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.ENUM, format=windowShade.Enums.windowShadeEnum)

        @dataclass
        class supportedWindowShadeCommands(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Window shade commands supported by this instance of Window Shade'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.LIST, format=List[windowShade.Enums.windowShadeEnum])

    class Functions:
        @dataclass
        class open(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Open the window shade'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.VOID

        @dataclass
        class close(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Close the window shade'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.VOID

        @dataclass
        class pause(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Pause opening or closing the window shade'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.VOID


@dataclass
class curtain(Skill):

    @MXClassproperty
    def descriptor(cls) -> str:
        return 'Allows for the control of the curtain.'

    class Enums:
        @MXClassproperty
        def descriptor(cls) -> str:
            return textwrap.dedent(
                '''\
                • closed - closed
                • closing - closing…
                • open - open
                • opening - opening…
                • partially open - partially open
                • paused -
                • unknown - unknown'''
            )

        class curtainEnum(MXStrEnum):
            closed = aauto()
            closing = aauto()
            open = aauto()
            opening = aauto()
            partially = aauto()
            paused = aauto()
            unknown = aauto()

    class Values:
        @dataclass
        class curtain(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'A string representation of whether the curtain is open or closed'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.ENUM, format=curtain.Enums.curtainEnum)

        @dataclass
        class supportedCurtainCommands(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Curtain commands supported by this instance of Curtain'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.LIST, format=List[curtain.Enums.curtainEnum])

    class Functions:
        @dataclass
        class open(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Open the curtain'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.VOID

        @dataclass
        class close(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Close the curtain'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.VOID

        @dataclass
        class pause(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Pause opening or closing the curtain'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.VOID


@dataclass
class blind(Skill):

    @MXClassproperty
    def descriptor(cls) -> str:
        return 'Allows for the control of the blind.'

    class Enums:
        @MXClassproperty
        def descriptor(cls) -> str:
            return textwrap.dedent(
                '''\
                • closed - closed
                • closing - closing…
                • open - open
                • opening - opening…
                • partially open - partially open
                • paused -
                • unknown - unknown'''
            )

        class blindEnum(MXStrEnum):
            closed = aauto()
            closing = aauto()
            open = aauto()
            opening = aauto()
            partially = aauto()
            paused = aauto()
            unknown = aauto()

    class Values:
        @dataclass
        class blind(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'A string representation of whether the blind is open or closed'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.ENUM, format=blind.Enums.blindEnum)

    class Functions:
        @dataclass
        class open(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Open the blind'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.VOID

        @dataclass
        class close(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Close the blind'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.VOID

        @dataclass
        class pause(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Pause opening or closing the blind'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.VOID


@dataclass
class robotCleanerCleaningMode(Skill):

    @MXClassproperty
    def descriptor(cls) -> str:
        return 'Allows for the control of the robot cleaner cleaning mode.'

    class Enums:
        @MXClassproperty
        def descriptor(cls) -> str:
            return textwrap.dedent(
                '''\
                • auto - The robot cleaner cleaning mode is in "auto" mode
                • part - The robot cleaner cleaning mode is in "part" mode
                • repeat - The robot cleaner cleaning mode is in "repeat" mode
                • manual - The robot cleaner cleaning mode is in "manual" mode
                • stop - The robot cleaner cleaning mode is in "stop" mode
                • map - The robot cleaner cleaning mode is in "map" mode'''
            )

        class robotCleanerCleaningModeEnum(MXStrEnum):
            auto = aauto()
            part = aauto()
            repeat = aauto()
            manual = aauto()
            stop = aauto()
            map = aauto()

    class Values:
        @dataclass
        class robotCleanerCleaningMode(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Current status of the robot cleaner cleaning mode'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.ENUM, format=robotCleanerCleaningMode.Enums.robotCleanerCleaningModeEnum)

    class Functions:
        @dataclass
        class setRobotCleanerCleaningMode(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Set the robot cleaner cleaning mode'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.VOID

            class Arguments:
                @dataclass
                class mode(SkillFunctionArgument):
                    @MXClassproperty
                    def descriptor(cls) -> str:
                        return 'Set the robot cleaner cleaning mode, to "auto", "part", "repeat", "manual" or "stop" modes'

                    @MXClassproperty
                    def argument_type(cls) -> MXValueType:
                        return MXValueType(MXType.ENUM, format=robotCleanerCleaningMode.Enums.robotCleanerCleaningModeEnum)


@dataclass
class powerMeter(Skill):

    @MXClassproperty
    def descriptor(cls) -> str:
        return 'Allows for reading the power consumption from devices that report it'

    class Values:
        @dataclass
        class power(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return (
                    'A number representing the current power consumption. Check the device documentation for how this value is reported - unit: Watts'
                )

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.DOUBLE)

        @dataclass
        class powerConsumption(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'energy and power consumption during specific time period - unit: Wh'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.DICT)


@dataclass
class presenceSensor(Skill):

    @MXClassproperty
    def descriptor(cls) -> str:
        return 'The ability to see the current status of a presence sensor device'

    class Enums:
        @MXClassproperty
        def descriptor(cls) -> str:
            return textwrap.dedent(
                '''\
                • present - The device is present
                • not present - left'''
            )

        class presenceEnum(MXStrEnum):
            present = aauto()
            not_present = aauto()

    class Values:
        @dataclass
        class presence(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'The current state of the presence sensor'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.ENUM, format=presenceSensor.Enums.presenceEnum)


@dataclass
class soundSensor(Skill):

    @MXClassproperty
    def descriptor(cls) -> str:
        return 'A Device that senses sound'

    class Enums:
        @MXClassproperty
        def descriptor(cls) -> str:
            return textwrap.dedent(
                '''\
                • detected - Sound is detected
                • not detected - no sound'''
            )

        class soundEnum(MXStrEnum):
            detected = aauto()
            not_detected = aauto()

    class Values:
        @dataclass
        class sound(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Whether or not sound was detected by the Device'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.ENUM, format=soundSensor.Enums.soundEnum)


@dataclass
class leakSensor(Skill):

    @MXClassproperty
    def descriptor(cls) -> str:
        return 'A Device that senses water leakage'

    class Enums:
        @MXClassproperty
        def descriptor(cls) -> str:
            return textwrap.dedent(
                '''\
                • detected - water leak is detected
                • not detected - no leak'''
            )

        class presenceEnum(MXStrEnum):
            detected = aauto()
            not_detected = aauto()

    class Values:
        @dataclass
        class leakage(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Whether or not water leakage was detected by the Device'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.ENUM, format=leakSensor.Enums.presenceEnum)


@dataclass
class atmosphericPressureSensor(Skill):

    @MXClassproperty
    def descriptor(cls) -> str:
        return 'Measure the atmospheric pressure'

    class Values:
        @dataclass
        class pressure(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Gets the value of the atmospheric pressure level.'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.DOUBLE)


@dataclass
class soundPressureLevel(Skill):

    @MXClassproperty
    def descriptor(cls) -> str:
        return 'Gets the value of the sound pressure level.'

    class Values:
        @dataclass
        class soundPressureLevel(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Level of the sound pressure'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.DOUBLE, [0, 194])


@dataclass
class smokeDetector(Skill):

    @MXClassproperty
    def descriptor(cls) -> str:
        return 'A device that detects the presence or absence of smoke.'

    class Enums:
        @MXClassproperty
        def descriptor(cls) -> str:
            return textwrap.dedent(
                '''\
                    • clear - No smoke detected
                    • detected - Smoke detected
                    • tested - Smoke detector test button was activated'''
            )

        class smokeEnum(MXStrEnum):
            clear = aauto()
            detected = aauto()
            tested = aauto()

    class Values:
        @dataclass
        class smoke(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'The state of the smoke detection device'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.ENUM, format=smokeDetector.Enums.smokeEnum)


@dataclass
class relativeHumidityMeasurement(Skill):

    @MXClassproperty
    def descriptor(cls) -> str:
        return 'Allow reading the relative humidity from devices that support it'

    class Values:
        @dataclass
        class humidity(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'A numerical representation of the relative humidity measurement taken by the device'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.DOUBLE, [0, 100])


@dataclass
class sirenMode(Skill):

    @MXClassproperty
    def descriptor(cls) -> str:
        return 'Allows for the control of the siren.'

    class Enums:
        @MXClassproperty
        def descriptor(cls) -> str:
            return textwrap.dedent(
                '''\
                • both - 
                • off - 
                • siren - 
                • strobe - '''
            )

        class sirenModeEnum(MXStrEnum):
            both = aauto()
            off = aauto()
            siren = aauto()
            strobe = aauto()

    class Values:
        @dataclass
        class sirenMode(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Current mode of the siren'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.ENUM, format=sirenMode.Enums.sirenModeEnum)

    class Functions:
        @dataclass
        class setSirenMode(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Set the siren mode'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.VOID

            class Arguments:
                @dataclass
                class mode(SkillFunctionArgument):
                    @MXClassproperty
                    def descriptor(cls) -> str:
                        return 'Set the siren mode'

                    @MXClassproperty
                    def argument_type(cls) -> MXValueType:
                        return MXValueType(MXType.ENUM, format=sirenMode.Enums.sirenModeEnum)


@dataclass
class windowControl(Skill):

    @MXClassproperty
    def descriptor(cls) -> str:
        return 'Allows for the control of the window shade.'

    class Enums:
        @MXClassproperty
        def descriptor(cls) -> str:
            return textwrap.dedent(
                '''\
                • closed - closed
                • open - open
                • unknown - unknown'''
            )

        class windowEnum(MXStrEnum):
            closed = aauto()
            open = aauto()
            unknown = aauto()

    class Values:
        @dataclass
        class window(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'A string representation of whether the window is open or closed'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.ENUM, format=windowControl.Enums.windowEnum)

    class Functions:
        @dataclass
        class open(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Open the window'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.VOID

        @dataclass
        class close(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Close the window'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.VOID


@dataclass
class weatherProvider(Skill):

    @MXClassproperty
    def descriptor(cls) -> str:
        return 'Provides weather information'

    class Enums:
        @MXClassproperty
        def descriptor(cls) -> str:
            return textwrap.dedent(
                '''\
                • thunderstorm - thunderstorm
                • drizzle - drizzle
                • rain - rain
                • snow - snow
                • mist - mist
                • smoke - smoke
                • haze - haze
                • dust - dust
                • fog - fog
                • sand - sand
                • ash - ash
                • squall - squall
                • tornado - tornado
                • clear - clear
                • clouds - clouds'''
            )

        class weatherEnum(MXStrEnum):
            thunderstorm = aauto()
            drizzle = aauto()
            rain = aauto()
            snow = aauto()
            mist = aauto()
            smoke = aauto()
            haze = aauto()
            dust = aauto()
            fog = aauto()
            sand = aauto()
            ash = aauto()
            squall = aauto()
            tornado = aauto()
            clear = aauto()
            clouds = aauto()

    class Values:
        @dataclass
        class weather(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Current weather condition'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.ENUM, format=weatherProvider.Enums.weatherEnum)

        class humidityWeather(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Current humidity level'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.DOUBLE)

        class pressureWeather(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Current pressure level'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.DOUBLE)

        class temperatureWeather(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Current temperature level'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.DOUBLE)

    class Functions:
        @dataclass
        class getWeatherInfo(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Get the current weather information - Return whole weather information, format: "temperature, humidity, pressure, pm25, pm10, weather, weather_string, icon_id, location"'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.STRING

            class Arguments:
                @dataclass
                class lat(SkillFunctionArgument):
                    @MXClassproperty
                    def descriptor(cls) -> str:
                        return 'The latitude of the location'

                    @MXClassproperty
                    def argument_type(cls) -> MXValueType:
                        return MXValueType(MXType.DOUBLE)

                @dataclass
                class lon(SkillFunctionArgument):
                    @MXClassproperty
                    def descriptor(cls) -> str:
                        return 'The longitude of the location'

                    @MXClassproperty
                    def argument_type(cls) -> MXValueType:
                        return MXValueType(MXType.DOUBLE)


@dataclass
class emailProvider(Skill):

    @MXClassproperty
    def descriptor(cls) -> str:
        return 'Provides email services'

    class Functions:
        @dataclass
        class sendMail(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Send an email'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.VOID

            class Arguments:
                @dataclass
                class toAddress(SkillFunctionArgument):
                    @MXClassproperty
                    def descriptor(cls) -> str:
                        return 'The email address of the recipient'

                    @MXClassproperty
                    def argument_type(cls) -> MXValueType:
                        return MXValueType(MXType.STRING)

                @dataclass
                class title(SkillFunctionArgument):
                    @MXClassproperty
                    def descriptor(cls) -> str:
                        return 'The title of the email'

                    @MXClassproperty
                    def argument_type(cls) -> MXValueType:
                        return MXValueType(MXType.STRING)

                @dataclass
                class text(SkillFunctionArgument):
                    @MXClassproperty
                    def descriptor(cls) -> str:
                        return 'The text of the email'

                    @MXClassproperty
                    def argument_type(cls) -> MXValueType:
                        return MXValueType(MXType.STRING)

        @dataclass
        class sendMailWithFile(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Send an email with an attachment'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.VOID

            class Arguments:
                @dataclass
                class toAddress(SkillFunctionArgument):
                    @MXClassproperty
                    def descriptor(cls) -> str:
                        return 'The email address of the recipient'

                    @MXClassproperty
                    def argument_type(cls) -> MXValueType:
                        return MXValueType(MXType.STRING)

                @dataclass
                class title(SkillFunctionArgument):
                    @MXClassproperty
                    def descriptor(cls) -> str:
                        return 'The title of the email'

                    @MXClassproperty
                    def argument_type(cls) -> MXValueType:
                        return MXValueType(MXType.STRING)

                @dataclass
                class text(SkillFunctionArgument):
                    @MXClassproperty
                    def descriptor(cls) -> str:
                        return 'The text of the email'

                    @MXClassproperty
                    def argument_type(cls) -> MXValueType:
                        return MXValueType(MXType.STRING)

                @dataclass
                class filePath(SkillFunctionArgument):
                    @MXClassproperty
                    def descriptor(cls) -> str:
                        return 'The path to the file to be attached'

                    @MXClassproperty
                    def argument_type(cls) -> MXValueType:
                        return MXValueType(MXType.STRING)


@dataclass
class menuProvider(Skill):

    @MXClassproperty
    def descriptor(cls) -> str:
        return 'Provides menu information services'

    class Functions:
        @dataclass
        class menu(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Get the menu - Return the menu list'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.STRING

            class Arguments:
                @dataclass
                class command(SkillFunctionArgument):
                    @MXClassproperty
                    def descriptor(cls) -> str:
                        return 'The command to get the menu - format: [오늘|내일] [학생식당|수의대식당|전망대(3식당)|예술계식당(아름드리)|기숙사식당|아워홈|동원관식당(113동)|웰스토리(220동)|투굿(공대간이식당)|자하연식당|301동식당] [아침|점심|저녁]'

                    @MXClassproperty
                    def argument_type(cls) -> MXValueType:
                        return MXValueType(MXType.STRING)

        @dataclass
        class todayMenu(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Get today\'s menu randomly - Return the menu list'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.STRING


@dataclass
class humidifierMode(Skill):

    @MXClassproperty
    def descriptor(cls) -> str:
        return 'Maintains and sets the state of an humidifier'

    class Enums:
        @MXClassproperty
        def descriptor(cls) -> str:
            return textwrap.dedent(
                '''\
                • auto -
                • low -
                • medium -
                • high -'''
            )

        class humidifierModeEnum(MXStrEnum):
            auto = aauto()
            low = aauto()
            medium = aauto()
            high = aauto()

    class Values:
        @dataclass
        class humidifierMode(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Current mode of the humidifier'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.ENUM, format=humidifierMode.Enums.humidifierModeEnum)

    class Functions:
        @dataclass
        class setHumidifierMode(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Set the humidifier mode'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.VOID

            class Arguments:
                @dataclass
                class mode(SkillFunctionArgument):
                    @MXClassproperty
                    def descriptor(cls) -> str:
                        return 'Set the humidifier mode to "auto", "low", "medium", or "high" mode'

                    @MXClassproperty
                    def argument_type(cls) -> MXValueType:
                        return MXValueType(MXType.ENUM, format=humidifierMode.Enums.humidifierModeEnum)


@dataclass
class pumpOperationMode(Skill):

    @MXClassproperty
    def descriptor(cls) -> str:
        return 'Allows for setting the operation mode on a pump.'

    class Enums:
        @MXClassproperty
        def descriptor(cls) -> str:
            return textwrap.dedent(
                '''\
                • normal - The pump is controlled by a setpoint.
                • minimum - This value sets the pump to run at the minimum possible speed it can without being stopped.
                • maximum - This value sets the pump to run at its maximum possible speed.
                • localSetting - This value sets the pump to run with the local settings of the pump.'''
            )

        class pumpOperationModeEnum(MXStrEnum):
            normal = aauto()
            minimum = aauto()
            maximum = aauto()
            localSetting = aauto()

    class Values:
        @dataclass
        class currentOperationMode(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'The current effective operation mode of the pump'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.ENUM, format=pumpOperationMode.Enums.pumpOperationModeEnum)

        @dataclass
        class operationMode(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'The operation mode of the pump'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.ENUM, format=pumpOperationMode.Enums.pumpOperationModeEnum)

        @dataclass
        class supportedOperationModes(SkillValue):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Supported operation modes for this device to be in'

            @MXClassproperty
            def value_type(cls) -> MXValueType:
                return MXValueType(MXType.LIST, format=List[pumpOperationMode.Enums.pumpOperationModeEnum])

    class Functions:
        @dataclass
        class setOperationMode(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Set the operation mode'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.VOID

            class Arguments:
                @dataclass
                class operationMode(SkillFunctionArgument):
                    @MXClassproperty
                    def descriptor(cls) -> str:
                        return 'The operation mode to set the device to'

                    @MXClassproperty
                    def argument_type(cls) -> MXValueType:
                        return MXValueType(MXType.ENUM, format=pumpOperationMode.Enums.pumpOperationModeEnum)


@dataclass
class calculator(Skill):

    @MXClassproperty
    def descriptor(cls) -> str:
        return 'Provides calculation services'

    class Functions:
        @dataclass
        class add(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Add two numbers'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.DOUBLE

            class Arguments:
                @dataclass
                class a(SkillFunctionArgument):
                    @MXClassproperty
                    def descriptor(cls) -> str:
                        return 'The first number to add'

                    @MXClassproperty
                    def argument_type(cls) -> MXValueType:
                        return MXValueType(MXType.DOUBLE)

                @dataclass
                class b(SkillFunctionArgument):
                    @MXClassproperty
                    def descriptor(cls) -> str:
                        return 'The second number to add'

                    @MXClassproperty
                    def argument_type(cls) -> MXValueType:
                        return MXValueType(MXType.DOUBLE)

        @dataclass
        class sub(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Subtract two numbers'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.DOUBLE

            class Arguments:
                @dataclass
                class a(SkillFunctionArgument):
                    @MXClassproperty
                    def descriptor(cls) -> str:
                        return 'The first number to add'

                    @MXClassproperty
                    def argument_type(cls) -> MXValueType:
                        return MXValueType(MXType.DOUBLE)

                @dataclass
                class b(SkillFunctionArgument):
                    @MXClassproperty
                    def descriptor(cls) -> str:
                        return 'The second number to add'

                    @MXClassproperty
                    def argument_type(cls) -> MXValueType:
                        return MXValueType(MXType.DOUBLE)

        @dataclass
        class mul(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Multiply two numbers'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.DOUBLE

            class Arguments:
                @dataclass
                class a(SkillFunctionArgument):
                    @MXClassproperty
                    def descriptor(cls) -> str:
                        return 'The first number to add'

                    @MXClassproperty
                    def argument_type(cls) -> MXValueType:
                        return MXValueType(MXType.DOUBLE)

                @dataclass
                class b(SkillFunctionArgument):
                    @MXClassproperty
                    def descriptor(cls) -> str:
                        return 'The second number to add'

                    @MXClassproperty
                    def argument_type(cls) -> MXValueType:
                        return MXValueType(MXType.DOUBLE)

        @dataclass
        class div(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Divide two numbers'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.DOUBLE

            class Arguments:
                @dataclass
                class a(SkillFunctionArgument):
                    @MXClassproperty
                    def descriptor(cls) -> str:
                        return 'The first number to add'

                    @MXClassproperty
                    def argument_type(cls) -> MXValueType:
                        return MXValueType(MXType.DOUBLE)

                @dataclass
                class b(SkillFunctionArgument):
                    @MXClassproperty
                    def descriptor(cls) -> str:
                        return 'The second number to add'

                    @MXClassproperty
                    def argument_type(cls) -> MXValueType:
                        return MXValueType(MXType.DOUBLE)

        @dataclass
        class mod(SkillFunction):
            @MXClassproperty
            def descriptor(cls) -> str:
                return 'Modulo two numbers'

            @MXClassproperty
            def return_type(cls) -> MXType:
                return MXType.DOUBLE

            class Arguments:
                @dataclass
                class a(SkillFunctionArgument):
                    @MXClassproperty
                    def descriptor(cls) -> str:
                        return 'The first number to add'

                    @MXClassproperty
                    def argument_type(cls) -> MXValueType:
                        return MXValueType(MXType.DOUBLE)

                @dataclass
                class b(SkillFunctionArgument):
                    @MXClassproperty
                    def descriptor(cls) -> str:
                        return 'The second number to add'

                    @MXClassproperty
                    def argument_type(cls) -> MXValueType:
                        return MXValueType(MXType.DOUBLE)
